
/* SSD1306Wire.h functions, see https://github.com/ThingPulse/esp8266-oled-ssd1306

  disply.clear()        - clear the local pixel buffer
  disply.display()      - write the buffer to the display
  disply.displayOff()   - turn the display OFF
  disply.displayOn()    - turn the display ON
  display.drawCircle(int16_t x, int16_t y, int16_t radius)
  display.drawFastImage(int16_t x, int16_t y, int16_t width, int16_t height, const uint8_t *image)
  display.drawHorizontalLine(int16_t x, int16_t y, int16_t length)
  display.drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1)  - draw from x0k,y0 to x1,y1
  display.drawProgressBar(uint16_t x, uint16_t y, uint16_t width, uint16_t height, uint8_t progress)
  display.drawRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.drawString(int16_t x, int16_t y, String text)
  display.drawStringMaxWidth(int16_t x, int16_t y, int16_t maxLineWidth, String text)
  display.drawVerticalLine(int16_t x, int16_t y, int16_t length)
  display.drawXbm(int16_t x, int16_t y, int16_t width, int16_t height, const char* xbm)
  display.fillCircle(int16_t x, int16_t y, int16_t radius)
  display.fillRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.flipScreenVertically()   - flip the current screen upside down
  disply.init()         - initialise the display
  disply.invertDisplay()- inverted display mode
  disply.end()          - free memory used by the display
  disply.displayOn()    - turn the display ON
  display.mirrorScreen() - draw the screen mirrored
  disply.normalDisplay()- normal display mode
  disply.reconnect()    - connect again through I2C
  display.setBrightness(uint8_t)
  display.setColor(OLEDDISPLAY_COLOR color) - BLACK, WHITE, INVERSE
  display.setContrast(uint8_t contrast, uint8_t precharge = 241, uint8_t comdetect = 64)
  display.setFont(const uint8_t* fontData)
  display.setPixel(int16_t x, int16_t y)  - set a pixel at x,y
  display.setTextAlignment(OLEDDISPLAY_TEXT_ALIGNMENT textAlignment)  - TEXT_ALIGN_LEFT, TEXT_ALIGN_CENTER, TEXT_ALIGN_RIGHT, TEXT_ALIGN_CENTER_BOTH

  W = display.getStringWidth(const char* text, uint16_t length)
  W = display.getStringWidth(String text)
  
  Display size is 128 x 64 (w x h) pixels
  Default fonts are ArialMT at 10, 16 & 24 pixels high
  Text options include:
    2 x 24, + 1 x 10 pnt lines
    1 x 24, + 2 x 16 pnt lines
    1 x 24, + 3 x 10 pnt lines
    4 x 16 pnt lines, no line spaces, decending characters will merge
    3 x 16, + 1 x 10 pnt lines
    6 x 10 pnt lines, no line spaces, decending characters will merge
*/

// --------------------------------------------------------------------------------

void Display_40ms() {
  // called every 40ms from the main loop
  // displays work as a menu system, level 0 to level options
  // if walking then simply display 'RoboDog'
  if (DispNOP) {return;}  // this function is aborted, no I2C activity
  if (Walk > 0) {
    // a walk movement has been initiated so simply display ROBODOG
    if (!DispWalk) {Display_ROBODOG();}
    return; // do nothing else
  } else {DispWalk = false;}
  
//  if (DispDel > 0) {DispDel--; return;} // don't display if DispDel > 0
  if (DispClr) {display.clear(); DispClr = false; return;}
  if (DispDisp) {display.display(); DispDisp = false; DispClr = true; return;}

  // we refresh individual displays based on a counter, DispCnt
  // set DispCnt -ve to prevent the display update process
  DispCnt--;
  if (DispCnt < 1) {
    // DispCnt is reset for display refresh, so can be different for each
    switch (MenuLvl) {
      case 0: // top level menu options
        switch (Menu0) {
          case Menu0_Battery:   DispCnt = 10; Display_Battery(); break;
          case Menu0_Demos:     DispCnt = 10; Display_DemoOff(); break;
          case Menu0_WiFiStats: DispCnt =  5; Display_WiFiStatus(); break;
          case Menu0_SpiritLvl: DispCnt =  4; Display_MPU_Levels(); break;
          case Menu0_LTOFRange: DispCnt =  5; Display_LTOF(); break;
          case Menu0_DispOff:   Display_Inactive(); break;
          default:
            Menu0 = 0; break;
        } break;
      case 1: // menu level 1
        switch (Menu0) {
          case Menu0_Battery: // battery menus
            switch (Menu1) {
              case 0: DispCnt = 10; Display_Battery(); break;
              case 1: DispCnt = 10; Display_BatData(); break;
              default: Menu1 = 0; break;
            } break;
          case Menu0_Demos: // Demo modes
            switch (Menu1) {
              case 0: DispCnt =  5; Display_Demo_LTOF(); break;
              case 1: DispCnt = 10; Display_Demo_Walk(); break;
              case 2: DispCnt = 10; Display_Demo_Balance(); break;
              case 3: DispCnt = 10; Display_Demo_Legs(); break;
              default: Menu1 = 0; break;
            } break;
          case Menu0_WiFiStats: // WiFi menus
            switch (Menu1) {
              case 0: DispCnt = 5; Display_WiFiStatus(); break;
              case 1: DispCnt = 5; Display_Wii_Ctrl(); break;
              default: Menu1 = 0; break;
            } break;
          case Menu0_SpiritLvl: // MUP6050
            switch (Menu1) {
              case 0: DispCnt = 4; Display_MPU_Levels(); break;
              case 1: DispCnt = 5; Display_MPU_Acc(); break;
              case 2: DispCnt = 5; Display_MPU_Gyr(); break;
              case 3: DispCnt = 5; Display_MPU_Ang(); break;
              default: Menu1 = 0; break;
            } break;
          case Menu0_LTOFRange: // LTOF
            Menu1 = 0;  // reset any button pressing
            DispCnt = 10; Display_LTOF(); break;
          default:  // handle the unexpected. This should never happen
            Menu0 = 0; break;
        }
        break;
      }
    DispDisp = true; // having created the display contents, set flag to display it

    // the following line puts a rectangle around every display, showing the boundry region
//    display.drawRect(0,0,128,64);
  }
}

// --------------------------------------------------------------------------------

void Display_Battery() {
  // display battery information
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 06, "Battery");
  if (BatAvg > BatCritical) {
    // battery powered mode
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(0, 27, String(BatAvg));
    display.setTextAlignment(TEXT_ALIGN_CENTER);
    display.drawString(64, 27, String(float(BatAvg)/BatCal) + "v");
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
    if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
    if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
    display.drawString(127, 27, String(AnyLong) + "%");
    // display the time
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(0, 48, getHmsTime(millis()));
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(127, 48, BatTime$);
  } else {
    // USB powered mode
    display.setTextAlignment(TEXT_ALIGN_CENTER);
    display.drawString(64, 27, "USB Pwr");
    // display the time
    display.drawString(64, 48, getHmsTime(millis()));
  }
}

// --------------------------------------------------------------------------------

void Display_BatData() {
  // plot the BatData[] array as a rolling average
  DispCX = 4;
  for (int zB = 0; zB < 60; zB++) {
    if (BatData[zB] > 0) {display.drawRect(DispCX,63 - BatData[zB],2,BatData[zB]);}
    DispCX += 2;
  }
  // draw axice
  DispCX = 4;
  display.drawLine(DispCX,0,DispCX,63);
  display.drawLine(DispCX,63,DispCX + 118,63);
}

// --------------------------------------------------------------------------------

void Display_Demo_Balance() {
  // menu 1 balance demo screen
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 2, "DEMO");
  display.drawString(64, 24, "Self-BALANCING");
  display.drawString(64, 48, "Press SW0");
}

// --------------------------------------------------------------------------------

void Display_Demo_Legs() {
  // menu 1 upside down legs demos
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 2, "DEMO");
  display.drawString(64, 24, "LEGS DEMO");
  if (UpRht) {
    display.drawString(64, 48, "Not upside Down");
  } else {
    display.drawString(64, 48, "Click SW0");
  }
}

// --------------------------------------------------------------------------------

void Display_Demo_LTOF() {
  // menu 1 LTOF demos
  if (LTOF_On > 0) {Display_LTOF(); return;}

  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 2, "DEMO");
  display.drawString(64, 24, "LTOF Range");
  display.drawString(64, 48, "Click SW0");
}

// --------------------------------------------------------------------------------

void Display_DemoOff() {
  // menu 0 DEMOS INACTIVE display  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 14, "DEMOS");
  display.drawString(64, 34, "INACTIVE");
}

// --------------------------------------------------------------------------------

void Display_Demo_Walk() {
  // menu 1 Walk demos
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 2, "DEMO");
  display.drawString(64, 24, "WALKING");
  if (AtRest) {
    display.drawString(64, 48, "Must be standing");
  } else {
    display.drawString(64, 48, "Click SW0");
  }
}

// --------------------------------------------------------------------------------

void Display_Held() {
  // called when locking a display mode, puts LOCKED in centre of screen but does
  // not clear the rest of the contents
  display.setFont(ArialMT_Plain_24);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  DispCX = display.getStringWidth("LOCKED");
  DispCY = 2 + (DispCX/2);
  display.setColor(BLACK);
  display.fillRect(64 - DispCY,21,DispCX + 5,25);
  display.setColor(WHITE);
  display.drawRect(64 - DispCY,21,DispCX + 5,25);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 20, "LOCKED");
  display.display();
  DispClr = true; // initiate a clear screen for the next update
}

// --------------------------------------------------------------------------------

void Display_Inactive() {
  // Display an inactive message, normally for use with graphing modes  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 14, "Inactive");
  display.drawString(64, 34, "Display");
  display.display();
  DispClr = true;  // display intro, then clear memory ready for first screen
  DispNOP = true;  // prevent future display updates unless this flag is cleared
}

// --------------------------------------------------------------------------------

void Display_Init() {
  // initialise the OLED display
  display.init();
  display.clear();
  display.setContrast(255);
  display.display();
  DispClr = false; DispDisp = false;
}

// --------------------------------------------------------------------------------

void Display_Intro() {
  // first screen to be displayed
  // default fonts are ArialMT_Plain_10, ArialMT_Plain_16 and ArialMT_Plain_24
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 6, "RoboDog");
  display.setFont(ArialMT_Plain_10);
  display.drawString(64, 28, "v0.00");
  display.drawString(64, 50, myMAC);
  display.display();
  DispClr = true;  // display intro, then clear memory ready for first screen
}

// --------------------------------------------------------------------------------

void Display_LTOF() {
  // display the range value from the LTOF if it is enabled
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);

  if (LTOF_On < 1) {
    // LTOF is not active
    display.drawString(64, 6, "LTOF");
    display.drawString(64,42, "Disabled");
  } else {
    // LTOF is active
    display.drawString(64, 0, "LTOF");
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    
    display.drawString(62, 16, "Range:");
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(66, 16, String(Range));
    
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(62, 32, "Filter:");
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(66, 32, String(RangeFtd));
    
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(62, 48, "FPS:");
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(66, 48, String(LTOF_fps));
  }
}

// --------------------------------------------------------------------------------

void Display_Menu_Level() {
  // called when changing menu mode, displays menu level without clearing
  // the rest of the contents
  display.setFont(ArialMT_Plain_24);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  DispCX = display.getStringWidth(" MENU " + String(MenuLvl) + " ");
  DispCY = 2 + (DispCX/2);
  display.setColor(BLACK);
  display.fillRect(64 - DispCY,21,DispCX + 5,25);
  display.setColor(WHITE);
  display.drawRect(64 - DispCY,21,DispCX + 5,25);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 20, " MENU " + String(MenuLvl) + " ");
  display.display();
  DispClr = true; // initiate a clear screen for the next update
}

// --------------------------------------------------------------------------------

void Display_MPU_Acc() {
  // display MPU6050 accelerometer information
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 0, "MPU Acc.");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 16, "AcX");
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(36, 16, String(AcX));
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 16, String(AcXAvg));

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 32, "AcY");
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(36, 32, String(AcY));
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 32, String(AcYAvg));

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 48, "AcZ");
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(36, 48, String(AcZ));
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 48, String(AcZAvg));
}

// --------------------------------------------------------------------------------

void Display_MPU_Ang() {
  // display MPU6050 angle information
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 0, "MPU Angles");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 16, "Pitch");
  display.drawString(40, 16, String(PitchGyr,1) + "°");
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 16, String(PitchAcc,1) + "°");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 32, "Yaw");
  display.drawString(40, 32, String(YawGyr,1) + "°");
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.drawString(127, 32, String(YawAcc,1) + "°");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 48, "Rot.");
  display.drawString(40, 48, String(RotGyr,1) + "°");
}

// --------------------------------------------------------------------------------

void Display_MPU_Gyr() {
  // display MPU6050 gyroscope information
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 0, "MPU Gyros");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(0, 16, "GyX");
  display.drawString(40, 16, String(GyX));

  display.drawString(0, 32, "GyY");
  display.drawString(40, 32, String(GyY));

  display.drawString(0, 48, "GyZ");
  display.drawString(40, 48, String(GyZ));

  if (GyCal < 1) {
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(127, 16, String(GyXAvg));
    display.drawString(127, 32, String(GyYAvg));
    display.drawString(127, 48, String(GyZAvg));
  }
}

// --------------------------------------------------------------------------------

void Display_MPU_Levels() {
  // display MPU6050 angles like s spirit level
  display.setFont(ArialMT_Plain_10);

  if ((fabs(PitchGyr) <= 2.0) && (fabs(YawGyr) <= 2.0)) {
    // angles are <= 2.0°
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(72, 0, String(PitchGyr,1) + "°");
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(24,27, String(YawGyr,1) + "°");
    display.drawLine(32,32,96,32);    // horizontal centre line
    DispCX = 62 + (16 * YawGyr);      // Yaw bubble
    display.drawRect(DispCX,30,5,5);  // horizontal bubble
    display.drawLine(64,0,64,63);     // vertical centre line
    DispCY = 30 - (16 * PitchGyr);    // Pitch bubble
    display.drawRect(62,DispCY,5,5);  // vertical bubble
  } else {
    // angles are <= 10.0°
    if (YawGyr >= 0.0) {
      DispCX = 96; DispCY = 2 * fabs(YawGyr);
      display.drawLine(DispCX,32 - DispCY,DispCX,32 + DispCY);  // horizontal wedge right
      display.drawLine(32,32,DispCX,32 - DispCY);
      display.drawLine(32,32,DispCX,32 + DispCY);
      display.setTextAlignment(TEXT_ALIGN_LEFT);
      display.drawString(DispCX + 4,27, String(YawGyr,1) + "°");
    } else {
      DispCX = 32; DispCY = 2 * fabs(YawGyr);
      display.drawLine(DispCX,32 - DispCY,DispCX,32 + DispCY);  // horizontal wedge left
      display.drawLine(96,32,DispCX,32 - DispCY);
      display.drawLine(96,32,DispCX,32 + DispCY);
      display.setTextAlignment(TEXT_ALIGN_RIGHT);
      display.drawString(DispCX - 4,27, String(YawGyr,1) + "°");
    }
    if (PitchGyr >= 0.0) {
      DispCX = 2 * fabs(PitchGyr); DispCY = 0; 
      display.drawLine(64 - DispCX,DispCY,64 + DispCX,DispCY);  // vertical wedge up
      display.drawLine(64,64,64 - DispCX,DispCY);
      display.drawLine(64,64,64 + DispCX,DispCY);
      if (YawGyr < 0.0) {
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(70,53, String(PitchGyr,1) + "°");
      } else {
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(58,53, String(PitchGyr,1) + "°");
      }
    } else {
      DispCX = 2 * fabs(PitchGyr); DispCY = 63; 
      display.drawLine(64 - DispCX,DispCY,64 + DispCX,DispCY);  // vertical wedge down
      display.drawLine(64,0,64 - DispCX,DispCY);
      display.drawLine(64,0,64 + DispCX,DispCY);
      if (YawGyr < 0.0) {
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(70,0, String(PitchGyr,1) + "°");
      } else {
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(58,0, String(PitchGyr,1) + "°");
      }
    }
  }
}

// --------------------------------------------------------------------------------

void Display_Reboot() {
  // called when a soft reset is about to be initiated
  display.clear();
  display.setFont(ArialMT_Plain_24);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 10, "RESET");
  display.setFont(ArialMT_Plain_16);
  display.drawString(64, 40, "Pending...");
  display.display();
}

// --------------------------------------------------------------------------------

void Display_ROBODOG() {
  // this is displayed when the Walk Engine is active
  // if only draws this display once then stops any display I2C activity
  display.clear();
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 20, "RoboDog");
  display.display();
  DispWalk = true; DispClr = true;
}

// --------------------------------------------------------------------------------

void Display_WiFiStatus() {
  // displays WiFi link data  
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 0, "WiFi STATUS");

  if (WiFiConnected) {
    if (WiFiRxBytes > 0) {
      // some data has been received
      display.setFont(ArialMT_Plain_10);
      display.setTextAlignment(TEXT_ALIGN_RIGHT);
      display.drawString(61, 18, "Rx Bytes:");
      display.setTextAlignment(TEXT_ALIGN_LEFT);
      display.drawString(64, 18, getNumCSV(WiFiRxBytes));
      display.setTextAlignment(TEXT_ALIGN_RIGHT);
      display.drawString(61, 30, "Rx Error:");
      display.setTextAlignment(TEXT_ALIGN_LEFT);
      display.drawString(64, 30, String(WiFiRxErr));
  
      display.setTextAlignment(TEXT_ALIGN_RIGHT);
      display.drawString(61, 41, "Tx Bytes:");
      display.setTextAlignment(TEXT_ALIGN_LEFT);
      display.drawString(64, 41, getNumCSV(WiFiTxBytes));
      display.setTextAlignment(TEXT_ALIGN_RIGHT);
      display.drawString(61, 53, "Tx Error:");
      display.setTextAlignment(TEXT_ALIGN_LEFT);
      display.drawString(64, 53, String(WiFiTxErr));
    } else {
      // transmission errors so display disconnected
      display.setTextAlignment(TEXT_ALIGN_CENTER);
      display.drawString(64, 22, "NO LINK");
      display.setFont(ArialMT_Plain_10);
      display.drawString(64, 48, String(myMAC));
    }
  } else {
    display.drawString(64, 22, "Disconnected");
    display.drawString(64, 44, "Trying " + String(WiFiTryNum));
  }
}

// --------------------------------------------------------------------------------

void Display_Wii_Ctrl() {
  // display 6 bytes received from Wii over ESP-NOW
  display.clear();
  // display I2C values as decimal 0 - 255
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.setFont(ArialMT_Plain_10); DispCX = 58;
  display.drawString(DispCX,  0, String(RxWiFi[0]));
  display.drawString(DispCX, 10, String(RxWiFi[1]));
  display.drawString(DispCX, 20, String(RxWiFi[2]));
  display.drawString(DispCX, 30, String(RxWiFi[3]));
  display.drawString(DispCX, 40, String(RxWiFi[4]));
  display.drawString(DispCX, 50, String(RxWiFi[5]));
  // display I2C values as binary 00000000 - 11111111
  display.setTextAlignment(TEXT_ALIGN_LEFT); DispCX = 66;
  display.drawString(DispCX,  0, GetBIN(RxWiFi[0]));
  display.drawString(DispCX, 10, GetBIN(RxWiFi[1]));
  display.drawString(DispCX, 20, GetBIN(RxWiFi[2]));
  display.drawString(DispCX, 30, GetBIN(RxWiFi[3]));
  display.drawString(DispCX, 40, GetBIN(RxWiFi[4]));
  display.drawString(DispCX, 50, GetBIN(RxWiFi[5]));
  // display 'I2C' in top left corner
  display.setFont(ArialMT_Plain_16);
  display.drawString(2, 0, "Wii");
       if (WiiType == 'C') {display.drawString(2, 18, "Csc");}
  else if (WiiType == 'N') {display.drawString(2, 18, "Nnk");}
}

// --------------------------------------------------------------------------------
