  
// ---------------------------------------------------------------------------------

void AddToMem() {
  // a comma has been received on the serial port, so load cmdVal into memory if
  // a memory load command has initiated this sequence of events
  // the memory must be filled from the bottom up
  if (MemRow < 0) {cmdVal = 0; return;}
  cmdVal *= cmdSgn; // correct cmdVal for -ve numbers
  if (MemPnt < MemWidth) {
    // only add values if the pointer is less than the memory row width, otherwise
    // the process wraps onto the next row
    Mem[MemRow + MemPnt] = cmdVal;
    if (GR == 0) {PrintTx += String(cmdVal) + ",";}
    MemPnt++;
    if (SI > 0) {
      // in SI mode apply angles directly to servos as recevied
      if (cmdVal < 0) {cmdVal = 0;} else if (cmdVal > 180) {cmdVal = 180;}
      setServoAng(cmdVal,ChMap[MemPnt - 1]);
      if (MemPnt == SI) {
        // data has filled row 0 servos so load them immediately
        if (GR == 0) {PrintTx += "\n";} // add a line feed to print buffer
        SI = 0; MemPnt = 0;   // reset pointers
      }
    } else if (MemPnt == MemWidth) {
      // data has filled the row
      if (GR == 0) {PrintTx += "\n";} // add a line feed to print buffer
      MemRow++; MemPnt = 0; // point at next row
      MemMax = (MemRow / MemWidth) + 1; // record the row filled as the max row number
//      PrintTx += "MemMax " + String(MemMax) + "\n";
    }
  }
  // zero cmdVal for receipt of next value
  cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1;
}

// ---------------------------------------------------------------------------------

void AutoPwrOFF() {
  // called every 10ms from the main loop
  // if no activity during time-out return to squat position and turn servo power off
  if (AutoOffTimer > 0) {
    AutoOffTimer--;
//    PrintTx += String(AutoOffTimer) + "\n";
    if (AutoOffTimer < 1) {
      if (!AtRest && (MainTask == 0)) {
        // not squatting and not doing anything important
        SetMainTask(98);  // move to squat position
        if (GR == 0) {PrintTx += "Auto Power OFF\n";}
      }
    }
  }
}

// ---------------------------------------------------------------------------------

void AvgAccAll() {
  // called as a 40ms subtask, keeps a 1 second rolling average of the acceleromters
  AcXSum += AcX; AcYSum += AcY; AcZSum += AcZ;
  if (AcAvgDiv < AcAvgDivMax) {
    // increase the scope of the averaging factor if necessary
    // note at 40ms sample, AvgDiv = 25 == 1 second rolling average
    AcAvgDiv++;
  } else {
    // once the accumulator has the correct number of samples roll out the average
    AcXSum -= AcXAvg; AcYSum -= AcYAvg; AcZSum -= AcZAvg;
  }
  AcXAvg = AcXSum/AcAvgDiv;
  AcYAvg = AcYSum/AcAvgDiv;
  AcZAvg = AcZSum/AcAvgDiv;
 }

// ---------------------------------------------------------------------------------

void AvgGyrAll() {
  // called as a 40ms subtask, keeps a rolling average of the gyroscopes
  GyXSum += GyX; GyYSum += GyY; GyZSum += GyZ;
  if (GyAvgDiv < GyAvgDivMax) {
    // increase the scope of the averaging factor if necessary
    // note at 40ms sample, AvgDiv = 25 == 1 second rolling average
    GyAvgDiv++;
  } else {
    // once the accumulator has the correct number of samples roll out the average
    GyXSum -= GyXAvg; GyYSum -= GyYAvg; GyZSum -= GyZAvg;
  }

  GyXAvg = GyXSum/GyAvgDiv;
  GyYAvg = GyYSum/GyAvgDiv;
  GyZAvg = GyZSum/GyAvgDiv;

  // initially the Gyro is not calibrated, so we run a timer before using its output
  // as an offset value
  if (GyCal > 0) {
    GyCal--; 
    if (GyCal < 1) {
      // calibration period complete so use offsets determined by averaging
      GyXOff = -GyXAvg;
      GyYOff = -GyYAvg;
      GyZOff = -GyZAvg;
      PitchGyr = PitchAcc;  // set pitch gyro using Acc angle
      RotGyr = 0.0;         // zero the rotation gyro
      YawGyr = YawAcc;      // set yaw gyro using Acc angle
    }
  }
 }

// ---------------------------------------------------------------------------------

void calcAngles() {
  // called from main loop every 40ms to calculate the yaw, pitch and turn angles
  PitchAcc = 57.2958 * asin((float)AcY/8192.0); // angle in radians
  YawAcc = -57.2958 * asin((float)AcX/8192.0);   // angle in radians

  // only do these calculations once the calibration process is completed
  if (GyCal > 0) {PitchGyr = 0.0; YawGyr = 0.0; RotGyr = 0.0; return;}

  // gyros are set at +/-250 deg/sec for 16-bit ADC = 32,767 FSD
  // to convert gyro value to deg/sec we divide by 131.068 (ie. 32,767/250)
  // we have recorded time period in GyTD in milliseconds, between readings
  // so angle is = GyAng * (GyTD/1000)/131.068
  PitchGyr -= (float)GyX * (float)GyTD / 131068.0;
  // use 1% accelerometer angle to correct any drift
  PitchGyr = PitchGyr + ((PitchAcc - PitchGyr) * 0.01);
  // limit the angle
  if (PitchGyr > 360.0) {PitchGyr -= 360.0;}
  else if (PitchGyr < -360.0) {PitchGyr += 360.0;}

  YawGyr -= (float)GyY * (float)GyTD / 131068.0;
  // use 1% accelerometer angle to correct any drift
  YawGyr = YawGyr + ((YawAcc - YawGyr) * 0.01);
  // limit the angle
  if (YawGyr > 360.0) {YawGyr -= 360.0;}
  else if (YawGyr < -360.0) {YawGyr += 360.0;}

  // perform long average offset adjustment to improve accuracy
  if (GyCal == 0) {
    if (PitchGyr < PitchAcc) {
      GyXOffCnt--;
      if (GyXOffCnt < -100) {GyXOffCnt = 0; GyXOff--;}
    } else {
      GyXOffCnt++;
      if (GyXOffCnt > 100) {GyXOffCnt = 0; GyXOff++;}
    }
    if (YawGyr < YawAcc) {
      GyYOffCnt--;
      if (GyYOffCnt < -100) {GyYOffCnt = 0; GyYOff--;}
    } else {
      GyYOffCnt++;
      if (GyYOffCnt > 100) {GyYOffCnt = 0; GyYOff++;}
    }
  }

  RotDif = (float)GyZ * (float)GyTD / 131068.0;
//  Serial.println(String(fabs(RotDif)) + "\t" + String(RotCnt));
  RotGyr += RotDif;
  if (fabs(RotDif) > 0.05) {RotCnt = 75;} else if (RotCnt == 0) {RotGyr = 0.0;}
  // switch the angle to zero after 3 sec time-out
  if (RotCnt > 0) {
    RotCnt--; if (RotCnt < 1) {
      RotGyr = 0.0; // zero the rotation gyro
    }
  }
  // limit the angle
  if (RotGyr > 360.0) {RotGyr -= 360.0;}
  else if (RotGyr < -360.0) {RotGyr += 360.0;}
}

// ---------------------------------------------------------------------------------

void Cal_ExportArrays() {
  // send the contents of the arrays to the serial port for the 16-Ch controller app
  // the general format is:
  // #F n
  // #L n,n,n,n,n,n,n,n,n,n,n,n,n,n,n,n,n
  // #R n,n,n,n,n,n,n,n,n,n,n,n,n,n,n,n,n
  // #U n,n,n,n,n,n,n,n,n,n,n,n,n,n,n,n,n
  int zV; // temporary pointer variable
  PrintTx += "#F " + String(pwmFreq) + "\n";
  PrintTx += "#L ";
  for (zV = 0; zV < 16; zV++) {
    PrintTx += String(SLL[zV]); if (zV < 15) {PrintTx += ",";}
  } PrintTx += "\n"; //PrintTxPacket(); delay(50);
  PrintTx += "#R ";
  for (zV = 0; zV < 16; zV++) {
    PrintTx += String(SRV[zV]); if (zV < 15) {PrintTx += ",";}
  } PrintTx += "\n"; //PrintTxPacket(); delay(50);
  PrintTx += "#U ";
  for (zV = 0; zV < 16; zV++) {
    PrintTx += String(SUL[zV]); if (zV < 15) {PrintTx += ",";}
  } PrintTx += "\n"; //next40ms = millis() + 50;
}

// ---------------------------------------------------------------------------------

void Cal_SetServoFreq() {
  // manually set servo PWM frequency, noiminally 50 Hz (40 - 100Hz)  
  if (cmdVal < 40) cmdVal = 40; // limit range values
  if (cmdVal > 100) cmdVal = 100; // limit range values
  if (cmdVal != pwmFreq) {
    pwmFreq = (float)cmdVal;
    I2Cpwm.setPWMFreq(pwmFreq);  // set PCA9685 PWM frequency
    PrintTx += "PWM Freq = " + String(pwmFreq) + "\n";
  }
}

// ---------------------------------------------------------------------------------

void Cal_SetServoOE(int zVal) {
  // manually set the servo board OE to HIGH or LOW
  if (zVal != 1) {
    // enabling PCA9685 outputs
    servoOE = 0;
    digitalWrite(Pin_OE, LOW); // enable all PWM outputs
    PrintTx += "OE=LO [ON]\n";
  } else {
    // disabling PCA9685 outputs. As this is asynchronous there is a risk that this
    // action will shorten one or more PWM pulses, so we remove them first
    servoOE = 1; ClearPCAop();
    delay(10 + (1000/int(pwmFreq)));  // wait longer than one PWM cycle
    digitalWrite(Pin_OE, HIGH); // disable all PWM outputs
    PrintTx += "OE=HI [OFF]\n";
  }
}

// ---------------------------------------------------------------------------------

void Cal_SetServoValue() {
  // set the target servo value LL - UL and report back
  if (cmdVal == 0) {
    // zero value sent from a disabled channel
    SRV[servoTgt] = cmdVal;
  } else if ((cmdVal >= int(SLL[servoTgt])) && (cmdVal <= int(SUL[servoTgt]))) {
    // in range value received
    SRV[servoTgt] = cmdVal;
  }
  I2Cpwm.setPWM(servoTgt, 0, SRV[servoTgt]);
  PrintTx += "SV=" + String(SRV[servoTgt]) + "\n";
}

// ---------------------------------------------------------------------------------

void Cal_SetServoVLL() {
  // store the lower limt value sent from the 16 channel servo controller
  SLL[servoTgt] = cmdVal;
}

// ---------------------------------------------------------------------------------

void Cal_SetServoVUL() {
  // store the upper limt value sent from the 16 channel servo controller
  SUL[servoTgt] = cmdVal;
}

// ---------------------------------------------------------------------------------

void clearBatData() {
  // called to crear the data array by filling it with min Bat values
  for (int zB = 0; zB < 60; zB++) {BatData[zB] = 0;}
}

// ---------------------------------------------------------------------------------

void ClearPCAop() {
  // sets all 16 PWM channels to zero pulses, effectively OFF
  // even with OE enabled the servos will not receive any pulses
  for (int zC=0; zC < 16; zC++) {
    I2Cpwm.setPWM(zC, 0, 0);  // if OFFtime == ONtime the op never goes HIGH
  }
}

// ---------------------------------------------------------------------------------

void clearSerOffsets() {
  // clear the servo offset array  
  for (int zS = 0; zS < 16; zS++) {SOF[zS] = 0;}
}

// ---------------------------------------------------------------------------------

void decodeKey(int zkeyVal) {
  // decodes zkeyVal and excutes commands on '.'
  // every time we receive a character set response to 10ms later, just in case
  // there is another packet behind it coming from the controller
  // when Ping data is received set AutoOffTimer and eye LEDs to red
  next40ms = millis() + 10;
  if (zkeyVal == 10) {return;}
  if (zkeyVal == 13) {return;}
  keyChar = char(zkeyVal);
  switch (keyChar) {
    case '.': doCmd(); return;          // command terminator
    case '+': cmdVal = -1; return;      // set flag for ML+. command
    case '-': cmdSgn = -1; return;      // numeric sign
    case ',': AddToMem(); return;       // assume cmdVal contains a memory value
    case '!': SetMainTask(98); return;  // return to 'Rest' position
    case '#': Ping = 50; AutoOffTimer = 0;  LEDTaskPnt = 3; return;   // 'ping' sent from application, disables auto-rest function
    case '*': cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1; return; // ESC abort command
    case '0': extendCmdVal(0); return;
    case '1': extendCmdVal(1); return;
    case '2': extendCmdVal(2); return;
    case '3': extendCmdVal(3); return;
    case '4': extendCmdVal(4); return;
    case '5': extendCmdVal(5); return;
    case '6': extendCmdVal(6); return;
    case '7': extendCmdVal(7); return;
    case '8': extendCmdVal(8); return;
    case '9': extendCmdVal(9); return;
  }
  if (cmdMode == ' ') {
    // test for new Command Mode char?
    // only accept certain characters as valid commands, or ignore them
    // delay battery update by 10 seconds whilst receiving commands
    BatRep = millis() + 10000;
    switch (keyChar) {
      // check for lower-case and convert
      case 'd': cmdMode = 'D'; break;
      case 'D': cmdMode = 'D'; break;
      case 'i': cmdMode = 'I'; break;
      case 'I': cmdMode = 'I'; break;
      case 'm': cmdMode = 'M'; break;
      case 'M': cmdMode = 'M'; break;
      case 'p': cmdMode = 'P'; break;
      case 'P': cmdMode = 'P'; break;
      case 'r': cmdMode = 'R'; break;
      case 'R': cmdMode = 'R'; break;
      case 's': cmdMode = 'S'; break;
      case 'S': cmdMode = 'S'; break;
      case 'y': cmdMode = 'Y'; break;
      case 'Y': cmdMode = 'Y'; break;
    } cmdType = ' '; cmdVal = 0;
  } else {
    // test for Command Type char?
    cmdType = keyChar;
    switch (keyChar) {
      // check for lower-case and convert to upper-case
      case 'a': cmdType = 'A'; break;
      case 'c': cmdType = 'C'; break;
      case 'b': cmdType = 'B'; break;
      case 'd': cmdType = 'D'; break;
      case 'e': cmdType = 'E'; break;
      case 'f': cmdType = 'F'; break;
      case 'i': cmdType = 'I'; break;
      case 'l': cmdType = 'L'; break;
      case 'o': cmdType = 'O'; break;
      case 'p': cmdType = 'P'; break;
      case 'r': cmdType = 'R'; break;
      case 's': cmdType = 'S'; break;
      case 't': cmdType = 'T'; break;
      case 'u': cmdType = 'U'; break;
      case 'v': cmdType = 'V'; break;
      case 'x': cmdType = 'X'; break;
      case 'z': cmdType = 'Z'; break;
    }
  }
}

// ---------------------------------------------------------------------------------

void disableWiFi() {
  // when either a switch is pressed or WiFi disable is required
  WiFiEn = false; WiFiCntC = 0; WiFiRestCnt = 0; C_Dn = 8; Z_Dn = 8;
}

// ---------------------------------------------------------------------------------

void doCmd() {
  // a '.' has been received on the serial port so execute command if valid
  // Commands:
  // Dn.        - set PID I-gain enable, 0 == OFF, 1 == ON
  // DD.        - decrease the I-gain by 1.0
  // DU.        - increase the I-gain by 1.0
  // DVnn.      - set 'I' gain value to nn/10
  // In.        - set PID I-gain enable, 0 == OFF, 1 == ON
  // ID.        - decrease the I-gain by 1.0
  // IU.        - increase the I-gain by 1.0
  // IVnn.      - set 'I' gain value to nn/10
  // MC.        - clears memory contents
  // ME.        - export the contents of the memory
  // MFnn.      - first row to be played in a memory sequence, default = 0
  // MLnn.n,n,. - start a memory load sequence, values separate by a comma, terminated by a comma
  // ML+.n,n,.  - continue memory load sequence
  // MO.        - run the contents of the memory once only.
  // MP.        - toggle walk pause if walking
  // MR.        - run the contents of the memory continuously.
  // MS.        - STOP a movement sequence
  // MXnn.      - run one row of the memory
  // Pn.        - set PID P-gain enable, 0 == OFF, 1 == ON
  // PBnn.      - set Pitch self-balance deadband value
  // PD.        - decrease the P-gain by 1.0
  // PMnn.      - set Pitch slef-balance Max value
  // PR.        - report PID gains and offsets
  // PSnn.      - Pitch self-balance value
  // PU.        - increase the P-gain by 1.0
  // PVnn.      - set 'P' gain value to nn/10
  // PZ.        - RESET PID gains and offsets
  // R.         - report XDebug data over the serial link
  // RL.        - reports XDebug lables for RPnn over the serial link
  // RPnn.      - report nn XDebug data over the serial link
  // RS.        - move to rest position.
  // SAnn.      - set target servo angle to nn immediately.
  // SE.        - export the calibration arrays to the serial port
  // SI.n,n,.   - set all twelve angles, values separate by a comma, terminated by a comma
  // SLnn.      - set the target servos lower limit
  // SOn.       - set PCA OE pin, 0 = enable, 1 = disable
  // SPnn.      - set leg speed to nn
  // SRn.       - set the state of the echo flag, 0 = OFF, 1 = ON
  // STnn.      - set the servo target channel for future angle data. Default = 0.
  // SUnn.      - set the target servos upper limit
  // SVnn.      - set the target servo PWM value
  // YBnn.      - set Yaw self-balance deadband value
  // YMnn.      - set Yaw slef-balance Max value
  // YSnn.      - Yaw self-balance value

  cmdVal*= cmdSgn;  // correct cmdVal with the sign value
  
  switch (cmdMode) {
    case ' ': break;
    case 'D':
      // PID controller D commands
      switch (cmdType) {
        case ' ': if (cmdVal == 1) {PID_d_En = true;} else {PID_d_En = false;} break;
        case 'D': pid_d_gain -= 1.0;  break;
        case 'U': pid_d_gain += 1.0; break;
        case 'V': pid_d_gain = (float)cmdVal/100.0; break;
      } PrintTx += "D" + String(cmdType) + String(cmdVal) + "\n"; break;
    case 'I':
      // PID controller I commands
      switch (cmdType) {
        case ' ': if (cmdVal == 1) {PID_i_En = true;} else {PID_i_En = false;} break;
        case 'D': pid_i_gain -= 1.0;  break;
        case 'U': pid_i_gain += 1.0; break;
        case 'V': pid_i_gain = (float)cmdVal/100.0; break;
      } PrintTx += "I" + String(cmdType) + String(cmdVal) + "\n"; break;
    case 'M':
      switch (cmdType) {
        case 'C': 
          // clears memory contents
          PrintTx += "MC.\n";
          clearMem(); break;
        case 'E': 
          // export the contents of the memory
          PrintTx += "ME.\n";
          exportMem(); break;
        case 'F':
          // change the start of a play sequence, default is 0
          Mem1st = cmdVal;
          PrintTx += "MF" + String(cmdVal) + ".\n"; break;
        case 'L': 
          // start a memory load sequence, values separate by a comma, terminated by a period '.'
          if (cmdVal >= 0) {
            // initiate a memory load at a specific memory location
            MemRow = MemWidth * cmdVal;
            PrintTx += "ML" + String(cmdVal) + ".";
          } else {
            // initiate a memory load at the next row
            MemRow += MemWidth;
            PrintTx += "ML+.";
          } MemPnt = 0; Trans = false; break;
        case 'O':
          // run the contents of the memory once only
          PrintTx += "MO.\n";
          SetMainTask(96); break; // run sequence from Mem[] once
        case 'P':
          if (Pause) {Pause = false;} // toggle pause flag
          else {Pause = true;}
          PrintTx += "MP.\n";
           break;
        case 'R':
          // run the contents of the memory continuously
          PrintTx += "MR.\n";
          SetMainTask(97); break; // run sequence from Mem[] continuously
        case 'S':
          // stop movements by terminating walk sequence loading process
          PrintTx += "STOP!!!\n"; setBalanceOFF();
          WalkMem = 0; break;
        case 'X':
          // move to the memory points Mem[MemRow]
          PrintTx += "MX" + String(cmdVal) + ".\n";
          MemRow = cmdVal; SetMainTask(95); break;
      } break;
    case 'P':
      // PID controller P commands
      switch (cmdType) {
        case ' ': if (cmdVal == 1) {PID_p_En = true;} else {PID_p_En = false;} break;
        case 'B': PitchGyrDb = (float)cmdVal/100.0; break;
        case 'D': pid_p_gain -= 1.0; break;
        case 'M': PitchSbMax = cmdVal; break;
        case 'R': Report_PID_Vals(); break;
        case 'S': setPitchSB(cmdVal); break;
        case 'U': pid_p_gain += 1.0; break;
        case 'V': pid_p_gain = (float)cmdVal/100.0; break;
        case 'Z': Reset_PID_Vals(); break;
      } PrintTx += "P" + String(cmdType) + String(cmdVal) + "\n"; break;
    case 'R':
      switch (cmdType) {
        case ' ': 
          // report the same set of values twice as WiFi data can easily be
          // corrupted. Therefore two consequitives messages must be received
          // for the PID Controller graphing function to work
          // the GR count is used to inhibit other forms of reporting, like battery
          RPT = 0; DebugReport(RPnn); DebugReport(RPnn); GR = 25; break;
        case 'L':
          // report trace labels for PID Graph app
          RPT = 1; DebugReport(RPnn); break;
        case 'P':
          RPnn = cmdVal; 
          PrintTx += "RP" + String(RPnn) + "\n"; break;
        case 'S':
          // move to rest position
          PrintTx += "RS.\n";
          SetMainTask(98); break;
      } break;
    case 'S':
      switch (cmdType) {
        case 'A':
          // set target servo angle to nn immediately
          PrintTx += "SA" + String(cmdVal) + ".\n";
          setServoAng(cmdVal,servoTgt);
          break;
        case 'E': Cal_ExportArrays(); break;
        case 'F': Cal_SetServoFreq(); break;
        case 'I':
          // set ten values immediately. Mem row 0 is used for this.
          PrintTx += "SI.";
          SI = MemWidth - 1; MemRow = 0; MemPnt = 0; break;
        case 'L': Cal_SetServoVLL(); break;
        case 'O': Cal_SetServoOE(cmdVal); break;
        case 'P':
          // set leg speed to cmdVal
          PrintTx += "SP" + String(cmdVal) + ".\n";
          Speed = cmdVal; setLegSpeed(cmdVal); break; // set a new leg speed
        case 'R':
          // set the state of the Echo flag
          if (cmdVal > 0) {Echo = true; PrintTx += "Echo ON\n";} else {Echo = false;}
          break;
        case 'T':
          // set the servo target channel for future angle data. Default = 0
          PrintTx += "ST" + String(cmdVal) + ".\n";
          servoTgt = cmdVal; break;
        case 'U': Cal_SetServoVUL(); break;
        case 'V': Cal_SetServoValue(); break;
      } break;
    case 'Y':
      switch (cmdType) {
        case 'B': YawGyrDb = (float)cmdVal/100.0; break;
        case 'M': YawSbMax = cmdVal; break;
        case 'S': setYawSB(cmdVal); break;
      } PrintTx += "Y" + String(cmdType) + String(cmdVal) + "\n"; break;
    default:
      PrintTx += "Unknown cmd: " + cmdMode + cmdType + String(cmdVal) + "\n";
      break;
  }
  // now reset the variables
  cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1;
}

// ---------------------------------------------------------------------------------

void exportMem() {
  // sends the contents of the move memory Mem[] to the serial port
  int zMemSize = MemRows * MemWidth;
  for (int zRow = 0; zRow < zMemSize; zRow+=11) {
    PrintTx += "\nML" + String(zRow/12) + ".";
    for (int zPnt = 0; zPnt < 11; zPnt++) {
      if (Mem[zRow + zPnt] < 99) {
        PrintTx += " ";
      }
      if (Mem[zRow + zPnt] < 12) {
        PrintTx += " ";
      }
      PrintTx += String(Mem[zRow + zPnt]) + ",";
    } Serial.flush(); Serial1.flush();
  } PrintTx += "\n";
}

// ---------------------------------------------------------------------------------

void extendCmdVal(int zVal) {
  // adds a new digit to the right-hand end of cmdVal
  cmdVal = abs(cmdVal * 10) + zVal;
}

// --------------------------------------------------------------------------------

void flashLEDs(int zCnt) {
  // flash the left and right LED swCnt times whilst running the main loop()
  // note the calling function must block being called from loop() to prevent stack
  // overflow conditions
  LED_Nop = true;  // block eye LED tasks
  for (int zP = 0; zP < zCnt; zP++) {
    NeoPixel_SetAllEyeLEDs(200, 0, 0); strip1.show();
    loopDelay(30);
    NeoPixel_SetAllEyeLEDs(0, 0, 0);strip1.show();
    loopDelay(270);
  }
  LED_Nop = false; // enable eye LED tasks
}

// -------------------------------------------------------------------------------

String get3Digit(int zV) {
  // returns a 3 character string from 0 - 999
  String zS = String(zV);
  if (zV < 100) {zS = "0" + zS;}
  if (zV < 10) {zS = "0" + zS;}
  return zS;
}

// --------------------------------------------------------------------------------

int getAngVal(int zAng) {
  // strips off any auto-rate offset and sets the temp AngRate flag
  // strips off any Pwr offset and sets the temp AngPwr flag
  if (zAng >= 400) {zAng -= 400; AngRate = true;} else {AngRate = false;}
  if (zAng >= 200) {zAng -= 200; AngPwr = false;} else {AngPwr = true;}
  return zAng;
}

// --------------------------------------------------------------------------------

String GetBIN(byte zVal) {
  // returns an 8 character binary representation of zVal
  String zBIN$ = "";
  // convert zVal into binary
  if (zVal & 0B10000000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B01000000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00100000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00010000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00001000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00000100) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00000010) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00000001) {zBIN$ += "1";} else {zBIN$ += "0";}
  return zBIN$;
}

// --------------------------------------------------------------------------------

String getHmTime(unsigned long zMillis) {
  // returns a 'time' string in Hrs:Min base on zMillis
  String zTime$ = ""; zMillis = zMillis/60000;  // reduce zMillis to minutes
  long zU = zMillis/60; zMillis = zMillis%60;    // get hours and minutes
  zTime$ = String(zU) + ":";
  if (zMillis < 10) {zTime$ += "0";}
  zTime$ += String(zMillis);  // add minutes
  return zTime$;
}

// --------------------------------------------------------------------------------

String getHmsTime(unsigned long zMillis) {
  // returns a 'time' string in Hrs:Min:sec base on zMillis
  String zTime$ = "";
  long zU = zMillis/3600000; zMillis = zMillis%3600000;
  if (zU > 0) {zTime$ = String(zU) + ":";}
  zU = zMillis/60000; zMillis = zMillis%60000;
  if (zU < 10) {zTime$ += "0";}
  zTime$ += String(zU);
  zU = zMillis/1000;
  if (zU < 10) {zTime$ += ":0";} else {zTime$ += ":";}
  zTime$ += String(zU);
  return zTime$;
}

// -------------------------------------------------------------------------------

void getLedCol(int zV) {
  // returns a colour for a specific zV
  switch (zV) {
    case 0: colR = 0; colG = 0; colB = 0; break;
    case 1: colR = 0; colG = 6; colB = 0; break;
    case 2: colR = 3; colG = 3; colB = 0; break;
    case 3: colR = 0; colG = 0; colB = 6; break;
    case 4: colR = 3; colG = 0; colB = 3; break;
    case 5: colR = 6; colG = 0; colB = 0; break;
  }
}

// -------------------------------------------------------------------------------

String getNumNNN(long zNum) {
  // returns a 3 digit string for numbers 000 - 999
  String zVal$ = String(zNum);
  if (zNum < 10) zVal$ = "00" + zVal$;
  else if (zNum < 100) zVal$ = "0" + zVal$;
  return zVal$;
}

// -------------------------------------------------------------------------------

String getNumCSV(long zNum ) {
  // returns a string of zNum with commas ie. 1,000,000
  String zVal$;
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000); zNum /= 1000;}
  else {zVal$ = String(zNum); return zVal$;}
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000) + zVal$; zNum /= 1000;}
  else {zVal$ = String(zNum) + zVal$; return zVal$;}
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000) + zVal$; zNum /= 1000;}
  else {zVal$ = String(zNum) + zVal$; return zVal$;}
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000) + zVal$; zNum /= 1000;}
  else {zVal$ = String(zNum) + zVal$; return zVal$;}
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000) + zVal$; zNum /= 1000;}
  else {zVal$ = String(zNum) + zVal$; return zVal$;}
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000) + zVal$; zNum /= 1000;}
  else {zVal$ = String(zNum) + zVal$; return zVal$;}
  return zVal$;
}

// -------------------------------------------------------------------------------

void getRangeCol(int zR) {
  // returns a colour for a range value
  // 0 = OFF
  // 1 = Green
  // 2 = Green/Yellow
  // 3 = Yellow
  // 4 = Orange
  // 5 = Red
  colR = 10; colG = 0; colB = 0;  // default is zR >= 5 Red
  switch (zR) {
    case 0: colR = 0; colG = 0; colB = 0; break;
    case 1: colR = 0; colG = 10; colB = 0; break;
    case 2: colR = 3; colG = 7; colB = 0; break;
    case 3: colR = 5; colG = 5; colB = 0; break;
    case 4: colR = 7; colG = 3; colB = 0; break;
  }
}

// -------------------------------------------------------------------------------

void Head_Tasks() {
  // called every 20ms to perform tasks based on a HeadTask pointer
  // an AtRest (Squat) condition will terminate current movement tasks
//  Serial.println(HeadTask);
  if ((HeadTask > 1) && AtRest) {HeadTask = 0;} // shut down tasks when going into squat

  switch(HeadTask) {
    case 0: //move head to centre forward and stop
      MoveHeadTo(90,100); HeadTask++; break;
    case 1: // default NOP state, performs auto-monitoring of dog activities
      break;

    case 10:  // move head to random positions
      MoveToRndHeadPos(50 + (6 * random(11)));  // random speed 50 to 110
      HeadTask++; break;
    case 11:  // wait for move to end
      if (Head == 0) {HeadTask++; HeadCnt = 20 + random(150);}
      break;
    case 12: // wait for random delay, then start the cycle over
      HeadCnt--; if (HeadCnt < 1) {HeadTask = 10;}
      break;
    case 13:  // stop the head from move at its current position
      Head = 0; break;

    case 20:  // scan head left/right and send data
      if (LTOF_On < 1) {VL53L0X_ON ();} // turn ON LTOF if it was OFF
      getHeadAngle(); LEDTaskPnt = 4;
      HdAngMin = 30; HdAngMax = 150; HeadInc = 4; Detect = false; DetectCnt = 0;
      HeadTask++; break;
    case 21:
      // report range and angle
      if (GR == 0) {PrintTx += "RA" + get3Digit(RangeFtd) + get3Digit(HeadAngle) + "\n";}
      if (RangeFtd <= 500) {
        DetectCnt = 3;
        if (!Detect) {
          // object just detected, so adjust limits
          if (HeadInc > 0) {HdAngMin = HeadAngle; HeadInc = 1;}
          else {HdAngMax = HeadAngle; HeadInc = -1;}
        } Detect = true;
      } else {Detect = false;}
      RangeNew = false; HeadTask++; break;
    case 22:
      // initiate a head move, and at the limits change direction
      HeadAngle += HeadInc;
      if (HeadAngle >= 150) {
        // reached upper limit so change direction
        HeadAngle = 150; DetectCnt = 0; HdAngMax = 150;
        HeadInc = -1; // change the sign of HeadInc
      } else if (HeadAngle <= 30) {
        // reached lower limit so change direction
        HeadAngle = 30; DetectCnt = 0; HdAngMin = 30;
        HeadInc = 1;  // change the sign of HeadInc
      } else {
        DetectCnt--;
        if (DetectCnt == 1) {
          // force a change in direction as we come out of the object detected zone
          if (HeadInc > 0) {HdAngMax = HeadAngle; HeadInc = -1;}  // we were increasing angle
          else {HdAngMin = HeadAngle; HeadInc = 1;} // we were decreasing angle
        } else if (DetectCnt == -12) {
          // an object has not been detected so revert back to normal speed after a delay
          HdAngMin = 30; HdAngMax = 150;
          if (HeadInc > 0) {HeadInc = 1;} else {HeadInc = -1;}
        }
      }
      // set the value of HeadInc based on its sign value
      if (abs(HeadInc) == 1) {HeadInc *= max((4 * (HdAngMax - HdAngMin))/120,1);}
      
//      MoveHeadTo(HeadAngle,100);
      setServoAng(HeadAngle,SH0); // move the head immediately to the new angle
      HeadTask++; break;
    case 23:
      // wait for new range
      if (RangeNew) {HeadTask = 21;}  // new value received so report it
      break;
  }
}

// -------------------------------------------------------------------------------

void initialiseMPU() {
  // By default the MPU-6050 sleeps. So we have to wake it up.
  Wire.beginTransmission(MPU_addr); //Start communication with the address found during search.
  Wire.write(0x6B); //We want to write to the PWR_MGMT_1 register (6B hex)
  Wire.write(0x00); //Set the register bits as 00000000 to activate the gyro
  I2CError = Wire.endTransmission();
  if (I2CError == 0) {
    //Set the full scale of the gyro to +/- 250 degrees per second
    Wire.beginTransmission(MPU_addr); //Start communication with the address found during search.
    Wire.write(0x1B); //We want to write to the GYRO_CONFIG register (1B hex)
    Wire.write(0x00); //Set the register bits as 00000000 (250dps full scale)
    Wire.endTransmission();
    //Set the full scale of the accelerometer to +/- 4g.
    Wire.beginTransmission(MPU_addr); //Start communication with the address found during search.
    Wire.write(0x1C); //We want to write to the ACCEL_CONFIG register (1A hex)
    Wire.write(0x08); //Set the register bits as 00001000 (+/- 4g full scale range)
    Wire.endTransmission();
    //Set some filtering to improve the raw data.
    Wire.beginTransmission(MPU_addr); //Start communication with the address found during search
    Wire.write(0x1A); //We want to write to the CONFIG register (1A hex)
    Wire.write(0x03); //Set the register bits as 00000011 (Set Digital Low Pass Filter to ~43Hz)
    Wire.endTransmission();
    Serial.println("MPU Initialised"); 
  } else {
    Serial.println("MPU Initialisation Failed!"); 
    Serial.println("I2C Error = " + String(I2CError)); 
  }
}

// --------------------------------------------------------------------------------

void LED_Task() {
  // called from main loop every 10 ms to control eye LEDs
  if (LED_Nop) {return;}  // eye LED tasks are blocked
  
  if (!sw0State || !sw1State) {
    // if either switch pressed light eye LEDs
    NeoPixel_SetAllEyeLEDs(200, 0, 0);
    NeoPixel_SetAllMthLEDs(0,0,0);
    LEDMask = true;
  }else if (LEDMask) {
    // turns OFF eye LEDs for 100ms
    NeoPixel_SetAllEyeLEDs(0, 0, 0);
    if (LEDTaskPnt != 99) {LEDNextTask = LEDTaskPnt; LEDTaskPnt = 99;}
    LEDTaskWait = 100; LEDMask = false;
  } else {
    if (LEDTaskPnt < -1) {LEDTaskPnt = abs(LEDTaskPnt); LEDSubTask = 0;}
    if (AtRest) {
      // previously doing something, but now returning to the rest state
      if (WiFiEn) {
        // if WiFiEn == true then set eyes blue
        if (LEDTaskPnt != 2) {eyeB = 10; LEDTaskPnt = 2; LEDSubTask = 0; HeadTask = 10;}
      } else {
        if (LEDTaskPnt > 0) {LEDTaskPnt = -1;}  // turn OFF eyes in manual mode
      }
    }

//    Serial.println("LEDTaskPnt = " + String(LEDTaskPnt));
    switch(LEDTaskPnt) {
      case -1:
        // reset task which turns OFF the eye LEDs
        eyeB = 0; eyeG = 0; eyeR = 0;
        NeoPixel_SetAllEyeLEDs(eyeR,eyeG,eyeB);
        LEDTaskPnt++; LEDSubTask = 0; LEDMask = false; break;
      case 0: break;  // do nowt
      case 1: LED_Task1_DimUpDwn(); break;
      case 2: LED_Task2_BlinkWhenStop(); break;
      case 3: LED_Task3_RedIfPowerOn(); break;
      case 4: LED_Task4_Red(); break;
      case 5: LED_Task5_Gear(); break;
      case 99:
        // wait for a predefined number of 10ms intervals
        LEDTaskWait--; if (LEDTaskWait < 1) {LEDTaskPnt = LEDNextTask;} break;
      default:
        LEDTaskPnt = -1; break;  // handle the unexpected
    }
  }
}

// --------------------------------------------------------------------------------

void LED_Task1_DimUpDwn() {
  // dim eye and mouth white LEDs up and down
  LEDTaskWait = 50;         // set the task wait timer in 1ms steps
  LEDSubTask++;
  if (LEDSubTask > (10 + random(10))) {
    LEDSubTask = 0; NeoPixel_SetAllMthLEDs(0,0,0);
    LEDTaskWait = 50 + (100 * random(5));
  }
  NeoPixel_SetAllEyeLEDs(LEDSubTask, LEDSubTask, LEDSubTask);
  LEDNextTask = LEDTaskPnt; // return to this task
  LEDTaskPnt = 99;          // invoke the wait task
}

// --------------------------------------------------------------------------------

void LED_Task2_BlinkWhenStop() {
  // track head movement and blink when it stops
  // if stationary then also blink after a random delay
  switch (LEDSubTask) {
    case 0:
      LEDTaskWait = 200 + random(800);  // random blink delay 2 - 10s, used if stattionary
      if (!AtRest || WiFiEn) {NeoPixel_SetAllEyeLEDs(eyeR,eyeG,eyeB);}
      LEDSubTask++; break;
    case 1: 
      // wait for moves to start or for end of a random delay
      LEDTaskWait--;
      if ((Head) || (Walk) || (LEDTaskWait < 1)) {LEDSubTask++;}
      break;
    case 2: if ((!Head) && (!Walk)) {LEDSubTask++;} break; // wait for moves to stop
    case 3:
      LEDTaskWait = 20 + random(20); LEDNextTask = LEDTaskPnt; LEDTaskPnt = 99;
      LEDSubTask++; break;  // restart the whole process
    case 4:
      // blink, then restart the whole process
      NeoPixel_SetAllEyeLEDs(0,0,0);
      LEDTaskWait = 10 + random(5); LEDNextTask = LEDTaskPnt; LEDTaskPnt = 99;
      LEDSubTask = 0; break;
    default:
      LEDSubTask = 0; break;  // handle the unexpected
  }
}

// --------------------------------------------------------------------------------

void LED_Task3_RedIfPowerOn() {
  // set the eyes to dim red when power is ON during trainer connection
  eyeG = 0; eyeB = 0;
  if (Ping > 0) {eyeR = 1;}
  else {eyeR = 0; LEDTaskPnt = -1;} // drop task when power goes off
  NeoPixel_SetAllEyeLEDs(eyeR,eyeG,eyeB);
}

// --------------------------------------------------------------------------------

void LED_Task4_Red() {
  // set the eyes to dim red, normally when LTOF is on
   eyeG = 0; eyeB = 0;
  if (LTOF_On > 0) {eyeR = 5;} else {eyeR = 0;}
  NeoPixel_SetAllEyeLEDs(eyeR,eyeG,eyeB);
}

// --------------------------------------------------------------------------------

void LED_Task5_Gear() {
  // temporarily display the gear colour in the eyes
  // you must save the current task pointer before calling this task
  switch (LEDSubTask) {
    case 0:
      getLedCol(Gear); NeoPixel_SetAllEyeLEDs(colR,colG,colB);
      LEDTaskWait = 20;
      if (LEDNextTask <= 0) {LEDNextTask = -1;} // LEDs were OFF so switch them OFF again
      LEDSubTask++; break;
    case 1:
      LEDTaskWait--;
      if (LEDTaskWait < 1) {
        LEDTaskPnt = LEDNextTask;
        NeoPixel_SetAllEyeLEDs(eyeR,eyeG,eyeB);
      } break;
    default:
      LEDSubTask = 0; break;  // handle the unexpected
  }
}

// --------------------------------------------------------------------------------

void LTOF_Read_Tasks() {
  // called from the main loop when enabled to read the laser range finder
  switch(LTOF_On) {
    // this is the VL53L0X_readRangeContinuousMillimeters() function broken down
    // into a state machine which take measurements when LTOF_On > 0
    // use the VL53L0X_Enable/Disable functions to switch the laser ON/OFF
    case 0: break;
    case 1: 
      // Returns a range reading in millimeters when continuous mode is active
      // (readRangeSingleMillimeters() also calls this function after starting a
      // single-shot range measurement)
      startTimeout(); LTOF_On++; break;
    case 2:
      // This is the main loop where we are waiting for the measurement to be completed
      // so we can branch to a multitasking loop from here whilst waiting
      // note that readReg takes about 500 us so not a fast while loop
      if ((VL53L0X_readReg(RESULT_INTERRUPT_STATUS) & 0x07) == 0) {
        // if no interrupt flag then check to see if a timeout has occured
        if (checkTimeoutExpired()) {
          // a timeout has occured
          did_timeout = true;
//          return 65535;
          Serial.println(F("LTOF timeout error!"));
        }
      } else {
        // an interrupt has occured suggesting a valid measurement
//        t1 = millis(); Serial.println(t1 - t0); t0 = t1; 
       LTOF_Wait = millis() + 32;  // default is +28
       LTOF_On++;
      } break;
    case 3:
      // assumptions: Linearity Corrective Gain is 1000 (default);
      // fractional ranging is not enabled
      Range = VL53L0X_readReg16Bit(RESULT_RANGE_STATUS + 10);
      Status = (VL53L0X_readReg(RESULT_RANGE_STATUS) & 0x78) >> 3;
      RangeNew = true;  // flag availabiulity of new range, this must be cleared by monitoring code
      VL53L0X_writeReg(SYSTEM_INTERRUPT_CLEAR, 0x01);
      LTOF_t0 = LTOF_t1; LTOF_t1 = millis();  // set timers for fps
      LTOF_fps = 1000/(LTOF_t1 - LTOF_t0);
      // limit raw Range to 1200, ie 1.2m as it gets very noisy beyond this
      if (Range > 1200) {Range = 1200;}
//      Serial.println(Range);
      LTOF_On++; break;
    case 4:
      // apply filtering to raw Range measurement
      RangeFtd = Range;
      if (Status != 11) {RangeFtd = 999;}
      if (RangeFtd >= 999) {
        RangeFtd = 999;
        if ((RangeFtd - (int)RangeLast) > 200) {
          Pulse++;
          if (Pulse < 3) {RangeFtd = RangeLast;}
        } else {Pulse = 0;}
      } else {
        if (((int)RangeLast - RangeFtd) > 200) {
          Pulse++;
          if (Pulse < 3) {RangeFtd = RangeLast;}
        } else {Pulse = 0;}
        if (RangeFtd < 50) {RangeFtd = 50;} // set minimum distance
      }
      // apply a filtered range limit of 600
      RangeLast = RangeFtd; RangeFtd = min(RangeFtd,600);
//      Serial.print(min(Range,1000)); Serial.print(","); Serial.println(RangeFtd);
      LTOF_On++; break;
    case 5:
      // as the sensor only returns measurements every 30-32ms, there is no point in
      // spending time reading interrupt registers
      if (millis() >= LTOF_Wait) {LTOF_On = 1;}
      break;
  }
}
//
//// --------------------------------------------------------------------------------
//
//int mapFP(float zX,float zImin,float zImax,float zOmin,float zOmax) {
//  // used because map() does not work well
//  return ((zX-zImin)*(zOmax-zOmin)/(zImax-zImin))+zOmin;
//}
//
//// --------------------------------------------------------------------------------
//
//int mapInt(long zX,long zImin,long zImax,long zOmin,long zOmax) {
//  // used because map() does not work well
//  return ((zX-zImin)*(zOmax-zOmin)/(zImax-zImin))+zOmin;
//}

// --------------------------------------------------------------------------------

void MouthTask() {
  // called every 20ms from main loop to perform a range mouth LED of tasks
  // it is controlled by the value of a task pointer LEDMthTask as follows:
  //
  // -1   - exit task returning to the default condition
  //  0   - battery 'wink' every 2 seconds, default state
  // 10   - LED bar graph left to right, colour changing with LEDMthVal
  // 20+  - scrolling, depending on MoveState
  //        MoveState =  0 - stationary
  //        MoveState =  1 - moving forward
  //        MoveState =  2 - moving forward bearing to the right
  //        MoveState =  3 - turning right
  //        MoveState =  5 - moving backward
  //        MoveState =  7 - turning left
  //        MoveState =  8 - moving forward bearing to the left
  //        MoveState = 13 - stepping right
  //        MoveState = 17 - stepping left
  // 30   - head L/R colour scanning
//  Serial.println(LEDMthTask);
  switch(LEDMthTask) {
    case -1:
      // normal exit task, turns OFF LEDs and goes back to task 0
      NeoPixel_SetAllMthLEDs(0,0,0); LEDMthTask = 0; LEDMthVal = 0; break;
    case 0:
      // default task simply displays battery voltage every 2 seconds
      // there are 5 LEDs in the mouth, 0 - 4, left to right
      mouthCnt--; getLedCol(Gear);
      if (mouthCnt == 10) {NeoPixel_SetMthLED(map(BatAvg,BatCritical,BatMax,0,5), colR, colG, colB);}
      else if (mouthCnt == 8) {NeoPixel_SetAllMthLEDs(0,0,0);}
      else if (mouthCnt < 1) {mouthCnt = 100;} // reset 2 second timer
      break;
    
    case 10: // bar graph task
      // LEDMthVal 0 - 5 determines which LEDs are lit
      getRangeCol(LEDMthVal);
      for (int zP = 0; zP < 5; zP++) {
        if (LEDMthVal > zP) {NeoPixel_SetMthLED(zP,colR,colG,colB);}
        else {NeoPixel_SetMthLED(zP, 0, 0, 0);}
      } break;


    case 20:
      // movement patterns. Colour determined by Speed value, green to red.
      // To initiate set LEDMthTask = 20; MoveState will determine the mode
      NeoPixel_Clear_Array(); getRangeCol(map(Speed,SpdMin,SpdMax,1,5));
      colBlu[0] = colB; colGrn[0] = colG; colRed[0] = colR;
      LEDMthPnt = 0; LEDMthCnt = 1;
      LEDMthTask += 1;  // moving forward by default
      if (MoveState ==  3) {LEDMthTask = 23;} // turning right
      if (MoveState ==  7) {LEDMthTask = 22;} // turning left
      if (MoveState ==  5) {LEDMthTask = 24;} // moving backwards
      if (MoveState == 13) {LEDMthTask = 23;} // turning right
      if (MoveState == 17) {LEDMthTask = 22;} // turning left
      LEDMthNext = LEDMthTask;
      break;
    case 21:
      // rotate the LEDs for forward movement
      // we only need to specify the values that change
      switch(LEDMthPnt) {
        case 0: LP[0] = 1; LP[1] = 1; LP[2] = 1; LP[3] = 1; LP[4] = 1; break;
        case 1:                       LP[2] = 0;                       break;
        case 2:            LP[1] = 0;            LP[3] = 0;            break;
        case 3: LP[0] = 0; LP[1] = 1; LP[2] = 1; LP[3] = 1; LP[4] = 0; break;
      } LEDMthTask = 25; break;
    case 22:
      // rotate the LEDs for turning left movement
      switch(LEDMthPnt) {
        case 0: LP[0] = 0; LP[1] = 1; LP[2] = 1; LP[3] = 1; LP[4] = 0; break;
        case 1: LP[0] = 1;                       LP[3] = 0; LP[4] = 1; break;
        case 2:                       LP[2] = 0; LP[3] = 1;            break;
        case 3:            LP[1] = 0; LP[2] = 1;                       break;
      } LEDMthTask = 25; break;
    case 23:
      // rotate the LEDs for turning right movement
      switch(LEDMthPnt) {
        case 0: LP[0] = 0; LP[1] = 1; LP[2] = 1; LP[3] = 1; LP[4] = 0; break;
        case 1: LP[0] = 1; LP[1] = 0;                       LP[4] = 1; break;
        case 2:            LP[1] = 1; LP[2] = 0;                       break;
        case 3:                       LP[2] = 1; LP[3] = 0;            break;
      } LEDMthTask = 25; break;
    case 24:
      // rotate the LEDs for backward movement
      switch(LEDMthPnt) {
        case 0: LP[0] = 1; LP[1] = 1; LP[2] = 1; LP[3] = 1; LP[4] = 1; break;
        case 1: LP[0] = 0;                                  LP[4] = 0; break;
        case 2: LP[0] = 1; LP[1] = 0;            LP[3] = 0; LP[4] = 1; break;
        case 3:            LP[1] = 1; LP[2] = 0; LP[3] = 1;            break;
      } LEDMthTask = 25; break;
      break;
    case 25:
       LEDMthPnt++; if (LEDMthPnt > 3) {LEDMthPnt = 0;}
      for (int zP = 0; zP < 5; zP++) {
        strip0.setPixelColor(zP,strip0.Color(colRed[LP[zP]],colGrn[LP[zP]],colBlu[LP[zP]]));
      } LEDshow0 = true;
      LEDMthTask++; break;
    case 26:
      LEDMthCnt--; if (LEDMthCnt < 1) {LEDMthCnt = 6; LEDMthTask = LEDMthNext;}
      break;


      case 30:  // linked to the head scanning task display colour quadrants
        NeoPixel_SetAllMthLEDs(0,0,0);
        LEDMthTask++; break;
      case 31:
        LEDMthPnt = map(HeadAngle,30,150,4,0);
        getRangeCol(map(RangeFtd,0,600,5,0));
        strip0.setPixelColor(LEDMthPnt,strip0.Color(colR,colG,colB)); LEDshow0 = true;
        break;
  }

  if ((LEDMthTask >= 21) && (LEDMthTask <= 24)) {
    // when moving change the LEDs colour to match the Speed
    getRangeCol(map(Speed,SpdMin,SpdMax,1,6));
    colBlu[0] = colB; colGrn[0] = colG; colRed[0] = colR;
  }
}

// --------------------------------------------------------------------------------

void NeoPixel_Clear_Array() {
  // clears the contents of the colour array
  for (int zL = 0; zL < 5; zL++) {
    colBlu[zL] = 0; colGrn[zL] = 0; colRed[zL] = 0;
  }
}

// --------------------------------------------------------------------------------

void NeoPixel_Load_Array(byte zP) {
  // load the contents of the colour array into the mouth LEDs
  // as the colour array is twice the depth as the number of mouth LEDs you can
  // pre-load the array and then simply
  for (int zL = 0; zL < 5; zL++) {
    strip0.setPixelColor(zL,strip0.Color(colRed[zL + zP],colGrn[zL + zP],colBlu[zL + zP]));
  } LEDshow0 = true;
}

// --------------------------------------------------------------------------------

void NeoPixel_SetAllEyeLEDs(byte zR, byte zG, byte zB) {
  // set all eye LEDs to one colour
  strip1.setPixelColor(0,strip1.Color(zR,zG,zB));
  strip1.setPixelColor(1,strip1.Color(zR,zG,zB));
  LEDshow1 = true;
}

// --------------------------------------------------------------------------------

void NeoPixel_SetAllMthLEDs(byte zR, byte zG, byte zB) {
  // set all mouth LEDs to one colour
  for (int zP = 0; zP < ledNum0; zP++) {
    colBlu[zP] = zB; colGrn[zP] = zG; colRed[zP] = zR;
    strip0.setPixelColor(zP,strip0.Color(colRed[zP],colGrn[zP],colBlu[zP]));
  } LEDshow0 = true;
}

// --------------------------------------------------------------------------------

void NeoPixel_SetEyeLED(byte zPixNum, byte zR, byte zG, byte zB) {
  // set the colour of eye LED zPixNum.
  colBlu[zPixNum] = zB; colGrn[zPixNum] = zG; colRed[zPixNum] = zR;
  strip1.setPixelColor(zPixNum,strip1.Color(colRed[zPixNum],colGrn[zPixNum],colBlu[zPixNum]));
  LEDshow1 = true;
}

// --------------------------------------------------------------------------------

void NeoPixel_SetMthLED(byte zPixNum, byte zR, byte zG, byte zB) {
  // set the colour of mouth LED zPixNum.
  colBlu[zPixNum] = zB; colGrn[zPixNum] = zG; colRed[zPixNum] = zR;
  strip0.setPixelColor(zPixNum,strip0.Color(colRed[zPixNum],colGrn[zPixNum],colBlu[zPixNum]));
  LEDshow0 = true;
}

// ----------------------------------------------------------------------

//void PrintBinary(int zB) {
//  // print a byte value in binary
//  if (zB < 128) {Serial.print("0");}
//  if (zB < 64) {Serial.print("0");}
//  if (zB < 32) {Serial.print("0");}
//  if (zB < 16) {Serial.print("0");}
//  if (zB < 8) {Serial.print("0");}
//  if (zB < 4) {Serial.print("0");}
//  if (zB < 2) {Serial.print("0");}
//  Serial.print(zB,BIN);
//  Serial.print("  ");
//  Serial.println(zB);
//}

// -------------------------------------------------------------------------------

void readAccelAll() {
  // read the X,Y,Z-axis acceleration values from the MPU
  // in this robot AcX = yaw, AcY = pitch, AcZ = up/down
//  t0 = micros();  // t1-t0 average = 225µs, longer with I2C bus contention
  AcXL = AcX; AcYL = AcY; AcZL = AcZ; // track previous values
  Wire.beginTransmission(MPU_addr);
  Wire.write(0x3B);  // starting with register 0x3B (ACCEL_XOUT_H)
  Wire.endTransmission(false);
  Wire.requestFrom(MPU_addr,6,true);  // request 1 register read
  AcX = Wire.read()<<8|Wire.read();  // 0x3B (ACCEL_XOUT_H) & 0x3C (ACCEL_XOUT_L)
  AcY = Wire.read()<<8|Wire.read();  // 0x3D (ACCEL_YOUT_H) & 0x3E (ACCEL_YOUT_L)
  AcZ = Wire.read()<<8|Wire.read();  // 0x3F (ACCEL_ZOUT_H) & 0x40 (ACCEL_ZOUT_L)

  // apply calibration offsets to these raw values
  AcX += AcXOff;  // note AcXOff should be zero initially in order to determin its value
  AcY += AcYOff;  // note AcYOff should be zero initially in order to determin its value
  AcZ += AcZOff;  // note AcZOff should be zero initially in order to determin its value

  // limit values to avoid trig function problems
  if (AcY > 8192) {AcY = 8192;} else if (AcY < -8192) {AcY = -8192;}
  if (AcX > 8192) {AcX = 8192;} else if (AcX < -8192) {AcX = -8192;}
  if (AcZ > 8192) {AcZ = 8192;} else if (AcZ < -8192) {AcZ = -8192;}

  // average over two readings to reduce random noise
  AcX = (AcX + AcXL)/2; AcY = (AcY + AcYL)/2; AcZ = (AcZ + AcZL)/2;

  if (AcZ <= 0) {UpRht = true;} else {UpRht = false;} // set the upright flag
//  t1 = micros();
}

// -------------------------------------------------------------------------------

//void readAccelX() {
//  // read the X-axis acceleration value from the MPU
//  Wire.beginTransmission(MPU_addr);
//  Wire.write(0x3B);  // starting with register 0x3B (ACCEL_XOUT_H)
//  Wire.endTransmission(false);
//  Wire.requestFrom(MPU_addr,2,true);  // request 1 register read
//  AcX = Wire.read()<<8|Wire.read();  // 0x3B (ACCEL_XOUT_H) & 0x3C (ACCEL_XOUT_L)
//  PrintTx += "AcX Raw = " + String(AcX) + "\n";
//}

// -------------------------------------------------------------------------------

//void readAccelY() {
//  // read the Y-axis acceleration value from the MPU
//  // AcY = 8,192 at 1g. In linerar region divide by 138.5 to get degrees
//  Wire.beginTransmission(MPU_addr);
//  Wire.write(0x3D);  // starting with register 0x3D (ACCEL_YOUT_H)
//  I2CError = Wire.endTransmission(false);
//  if (I2CError == 0) {
//    Wire.requestFrom(MPU_addr,2,true);  // request 1 register read
//    if (Wire.available() > 1) {
//      AcY = Wire.read()<<8|Wire.read();  // 0x3D (ACCEL_YOUT_H) & 0x3E (ACCEL_YOUT_L)
//      PrintTx += "AcY Raw = " + String(AcY) + "\n";
//    }
//  }
//  // we apply an offset if needed, depending on how it has been mounted
//  // uncomment the next line to determine the raw value at rest, AcYrest
//  // then set AcYOff = -AcYrest to zero out the offset
//  return;
//  AcYsum += (long)AcY; AcYsum -= AcYaV;   // update the rolling accumulator
//  AcYaV = AcYsum/4;                       // calc the average value
////  if (Fn) {AcYaV = AcYsum/40;}            // greater average in special function mode
////  AcYaV = AcY;                          // don't calc the average value
//    PrintTx += "AcY Av = " + String(AcYaV) + "\n";
////  AcYF = ((float)((AcYaV + AcYOff)/69))/2.0;         // convert to an approximate angle
////  AcYsgn = true; // set sign flag as positive
////  if (AcYF < 0.0){AcYF = -AcYF; AcYsgn = false;}; // set sign flag as negative
//}

// -------------------------------------------------------------------------------

//void readAccelZ() {
//  // read the Z-axis acceleration value from the MPU
//  Wire.beginTransmission(MPU_addr);
//  Wire.write(0x3F);  // starting with register 0x3F (ACCEL_ZOUT_H)
//  Wire.endTransmission(false);
//  Wire.requestFrom(MPU_addr,2,true);  // request 1 register read
//  AcZ=Wire.read()<<8|Wire.read();  // 0x3F (ACCEL_ZOUT_H) & 0x40 (ACCEL_ZOUT_L)
//  AcZ=AcZ + AcZOff;
//}

// --------------------------------------------------------------------------------

void readBattery(int zM) {
  // called from the main loop every 40ms (25Hz)
  // read the voltage on the analog input and average it over ten counts
  // zM = 0 - simple read + average
  // zM = 1 - calc float and percentage
  BatVol = analogRead(BatPin); BatTms = millis(); // take the reading
  BatSum = BatSum + BatVol - BatAvg;
  BatAvg = BatSum/20;     // ADC is averaged over 20 readings to remove noise
  // uncomment the following line if you want to see readings on serial monitor
//  Serial.print(F("BattAv = ")); Serial.print(BattAv); Serial.print(F("\t")); Serial.print((5.0 * BattAv)/608.9); Serial.println(F("v"));
  
  //#############################################################################
  //
  // Battery life prediction code + data logger
  //
  //#############################################################################
  // BatCnt is set to 10 second interval to start with, to ignore initial readings
  if (!USBPwr) {
    if (BatDel > 0) {BatDel--;}
    else {
      BatDel = 25;  // use a 1 second timer for prediction updates & data logging
      if (BatAvg > BatAvgMax) {BatAvgMax = BatAvg;}
      BatData[BatDP] = (62L * (BatAvg - BatCritical))/(BatAvgMax - BatCritical);
//      Serial.println("BatData: " + String(BatData[BatDP]));
      if (BatDP < 59) {BatDP++;} else {shiftBatData();}
      
      if (BatT0 == 0) {
        // this is the first measurement after 10 second start delay
        // this values are the cornerstone of the prediction
        BatT0 = BatTms; BatV0 = BatAvg; BatV1 = BatAvg;
//        Serial.print("BatT0 "); Serial.println(BatT0);
      } else {
        if (BatAvg != BatV1) {
          // there has been a change in the battery voltage setting a new discharge point
          // and the current battery voltage is different, so avoid divide by zero error
          // recalculate battery life predicition base on new value and time
//          Serial.print("BatAvg "); Serial.println(BatAvg);
          BatV1 = BatAvg; // record the latest low point
          if  (BatV0 == BatV1) {BatV1--;} // avoid divide by zero error
          AnyLong = ((BatTms - BatT0)*(BatV1 - BatCritical))/(BatV0 - BatV1);
      //      Serial.print("AnyLong "); Serial.println(AnyLong);
          if (AnyLong > 0) {
            BatTime$ = getHmTime(AnyLong);
          }
        }
      }
    }  
  
  //##############################################################################
  //
  //  critical battery test
  //
  //##############################################################################
  // test for battery low condition, 683 == 6.6v
    if (BatAvg < BatCritical) {
      // battery has reached critical level so squat down and blink centre LED
      BatCritCnt++;
      if (BatCritCnt >= 100) {
        // low voltage sustained for 4 seconds or more, so we shut down
        // the delay ignores brown-outs caused by heavy servo movements
        NeoPixel_SetAllEyeLEDs(0,0,0); strip1.show(); // eye LEDs OFF
        NeoPixel_SetAllMthLEDs(0,0,0); strip0.show(); // mouth LEDs OFF
        MoveToSquat(); WaitWhileMove(); // goto sqaut and power-down servos
        while (true) {
          // blink mouth centre LED red forever
          delay(20); BatCnt--;
          if (BatCnt < 1) {
            BatCnt = 100; NeoPixel_SetMthLED(2, 0, 0, 0); strip0.show();
          }
          if (BatCnt == 5) {NeoPixel_SetMthLED(2, 1, 0, 0); strip0.show();}
          // continue to take readings in case this is a deliberate power-down
          BatVol = analogRead(BatPin);
          BatSum = BatSum + BatVol - BatAvg; BatAvg = BatSum /20;
          if ((digitalRead(sw0Pin) == LOW) && (BatAvg > (BatCritical + 10))) {break;}
          yield();
        }
        // user has pressed button during power-down
        // battery looks ok so revive
        setDefaults();
        synchLoopTimers();
      }
    } else {BatCritCnt = 0;}  // reset the 4 second timer
  }
  
//  if (Ping == 9) {Ping = 0; zM = 1;} // force a battery report every 10 pings
  if ((zM < 1) && (millis() < BatRep)) {return;} // exit if not a report measurement
  
  // send battery voltage and % to serial port
  BatRep = millis() + 10000; // update reading every 10 seconds
  if (GR > 0) {return;} // don't report when in graphing mode
  
  AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
  if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
  if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
  PrintTx += "Batt " + String(float(BatAvg)/BatCal) + "v " + String(AnyLong) + "%\n";
}

// -------------------------------------------------------------------------------

void readGyrosAll() {
  // read the X,Y,Z-axis gyroscope values from the MPU
  // in this robot GyX = pitch, GyY = yaw, GYZ = rotate round
  // note gyro values are in degrees/sec so we use time between measurements to 
  // determine the average displaced angle
//  t0 = micros();  // t1-t0 average = 315µs, longer with I2C bus contention
  GyXL = GyX; GyYL = GyY; GyZL = GyZ; // track previous values
  Wire.beginTransmission(MPU_addr);
  Wire.write(0x43);  // starting with register 0x43 (GYRO_XOUT_H)
  Wire.endTransmission(false);
  Wire.requestFrom(MPU_addr,6,true);  // request 1 register read
  GyX = Wire.read()<<8|Wire.read();  // 0x43 (GYRO_XOUT_H) & 0x44 (GYRO_XOUT_L)
  GyY = Wire.read()<<8|Wire.read();  // 0x45 (GYRO_YOUT_H) & 0x46 (GYRO_YOUT_L)
  GyZ = Wire.read()<<8|Wire.read();  // 0x47 (GYRO_ZOUT_H) & 0x48 (GYRO_ZOUT_L)
  if (GyT > 0) {
    GyTD = millis() - GyT;  // get the differential time between readings
    GyT = millis();
  } else {
    // 1st reading after reset
    GyT = millis();
    GyTD = 40; GyXL = GyX; GyYL = GyY; GyZL = GyZ;
  }

  // apply calibration offsets to these raw values
  GyX += GyXOff;  // note GyXOff should be zero initially in order to determin its value
  GyY += GyYOff;  // note GyYOff should be zero initially in order to determin its value
  GyZ += GyZOff;  // note GyZOff should be zero initially in order to determin its value

  // average over two readings to reduce random noise
  GyX = (GyX + GyXL)/2; GyY = (GyY + GyYL)/2; GyZ = (GyZ + GyZL)/2;
//  t1 = micros();
}

// --------------------------------------------------------------------------------

void readSerial() {
  // reads characters from the serial port and responds to commands
  // set PrintTgt to indicate source as serial port for responses
  keyVal = Serial.read();
  if (keyVal != -1) {PrintTgt = 0; decodeKey(keyVal);}
}

// --------------------------------------------------------------------------------

void readSW0() {
  // read the left button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  if (SW0_Nop) {return;}  // block this function whilst == true
  
  sw0State = digitalRead(sw0Pin);
  if ((sw0LastState == HIGH) && (sw0State == LOW)) {
    // SW0 has just been pressed down
    // if 'walking' go immediately to the STAND position
//    PrintTx += "SW0 Hi-LO\n";
    sw0DwnTime = 0; // restart button down time counter
    if (Walk > 0) {
      PrintTx += "SW0.\n"; ESC0 = true; SetMainTask(90);
    } else {
      sw0Cnt++;     // count on the falling edge
      sw0Timer = 0; // reset the timer
      if (AutoOffTimer > 0) {AutoOffTimer = 1000;}  // extend auto-off timer if button pressed
    }
  }
  if (ESC0 && !sw0State) {return;}
  sw0LastState = sw0State; // record current state
  ESC0 = false;
  if (sw0State == LOW) {sw0DwnTime++;}  // track button pressed time
  if (sw0Cnt > 0) {sw0Timer++;}
  if ((sw0Timer >= 100) && (sw0State == HIGH)) {
    // button released for 1 sec so assume valid button count
//    Serial.print(F("swCnt = ")); Serial.println(swCnt);
    SW0_Nop = true; flashLEDs(sw0Cnt); // indicate count by flashing
    SW0_Nop = false;
    if ((Menu0 == Menu0_Demos) && (MenuLvl == 1)) {
      // display is in menu Demos mode, so react accordingly
      if (sw0Cnt == 1) {
        // most Demos are single click start/stop
        switch (Menu1) {
          case 0: // LTOF static demo
            // static LTOF can be performed in any position
            if (LTOF_On < 1) {SetMainTask(1);}  // turn LTOF ON
            else {SetMainTask(2);}              // turn LTOF OFF
            break;
          case 1: // walking demo
            if (Stand) {
              SetMainTask(3); // perform walk demo
            } break;
          case 2: // balance demo
            if (Stand || Ready) {
              SetMainTask(5); // perform balance demo
            } break;
          case 3: // upside down legs demo
            if (!UpRht) {
              // only do this if the robot is on its back
              SetMainTask(4); // perform seemingly random rate demo
            } break;
        }
      }
    } else {
      // display is not in menu Demos mode
      if (sw0Cnt == 1) {
        // single button press
             if (AtRest) {SetMainTask(94);}  // move to sit low position
        else if (SitLow) {SetMainTask(93);}  // move to sit high position
        else if (SitHigh) {SetMainTask(92);} // move to stand position
        else {PrintTx += "SW0 Eh?\n";}
  //    if (sw0Cnt == 1) {SetMainTask(1);}  // toggle LTOF
  //    if (sw0Cnt == 2) {SetMainTask(2);}; // backaway from target mode
  //    if (sw0Cnt == 3) {SetMainTask(3);}; // track target mode
  //    if (sw0Cnt == 4) {SetMainTask(4);}; // head scanning mode
      }
    }
    sw0Cnt = 0; sw0Timer = 0; synchLoopTimers();
  } else if((sw0Timer >= 200) && (sw0State == LOW)) {
    // button held down for >= 2 seconds
    if (!AtRest) {SetMainTask(98);} // move to squat position
    else {SetMainTask(91);}         // move to standing position
    sw0Cnt = 0; sw0Timer = 0;       // LedMode = 0;
  }
}

// --------------------------------------------------------------------------------

void readSW1() {
  // read the right button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  if (SW1_Nop) {return;}  // block this function whilst == true
  
  sw1State = digitalRead(sw1Pin);
  if (sw1State == LOW) {
    //##############################################################################
    //
    // SW1 button is pressed down
    //
    //##############################################################################
    if (DispCnt < 10) {DispCnt = 10;} // don't allow display to change when button is pressed
    if (sw1LastState == HIGH) {
      // SW1 has just been pressed down
      if (Walk > 0) {
        PrintTx += "SW1.\n"; ESC1 = true; SetMainTask(90);
      } else {
        sw1Cnt++;       // count on the falling edge
        sw1Timer = 0;   // reset the timer
        if (AutoOffTimer > 0) {AutoOffTimer = 1000;}  // extend auto-off timer if button pressed
      }
//    if (MainTask) {sw1Cnt = 0; SetMainTask(98); ESC1 = true;}
    }
    if (ESC1 && !sw1State) {return;}
    ESC1 = false;
    if (sw1Timer == 60) {
      // button held down for 600ms so toggle the display mode
           if (MenuLvl == 0) {MenuLvl = 1;}
      else if (MenuLvl == 1) {MenuLvl = 0;}
      else {MenuLvl = 0;}
      Display_Menu_Level();
    } else if (sw1Timer >= 200) {
      // button held down for >= 2 seconds so force a soft reset
      Serial.println("Reset"); Display_Reboot();
      Reboot = true;  // stop all main loop tasks
      resetFunc();    //call reset
    }
  } else {
    //##############################################################################
    //
    // sw1State == HIGH as SW1 button is not pressed
    //
    //##############################################################################
    if (sw1LastState == LOW) {
      // SW1 has just been released
      if ((sw1Timer > 0) && (sw1Timer < 60)) {
//        PrintTx += "sw1Timer = " + String(sw1Timer) + "\n";
        if (sw1Timer < 30) {
          // short press so increment the display
               if (MenuLvl == 0) {Menu0++;}
          else if (MenuLvl == 1) {Menu1++;}
        } else {
          // extended press so decrement the display
               if (MenuLvl == 0) {if (Menu0 > 0) {Menu0--;}}
          else if (MenuLvl == 1) {if (Menu1 > 0) {Menu1--;}}
        }
        DispNOP = false; DispClr = true; DispDisp = false; DispCnt = 0;
        sw1Cnt = 0; sw1Timer = 0;
      }
    }
    if (sw1Timer >= 100) {
      // button released for 1 sec so assume valid button count
      // we can put code here which responds to button counts
      sw1Cnt = 0; sw1Timer = 0;
    }
  }
  sw1LastState = sw1State; // record current state
  if (sw1Cnt > 0) {sw1Timer++;}
}

// ---------------------------------------------------------------------------------

void readWiFiRx() {
  // called from the main loop every cycle to check for ESP-NOW data  
  // a state machine is used to separate the data blocks by header type
  DispDel = 500;  // hold off display updates for 10 sec
  switch(Rx_Task) {
    case 0: // determine the type of data to handle
           if (Rx_Buff.ESPdata[Rx_Pnt] == 'B') {Rx_Task = 3; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'C') {WiiType = 'C'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'N') {WiiType = 'N'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '-') {WiiType = '-'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '$') {Rx_Task = 2; break;}
      // oh dear! bloick header not recognised so discard the block and flag error
      else {WiFiRxErr++; WiFiRx = false;}
      break;
      
    case 1: // load Wii I2C data
      // after the header byte there are 6 data bytes + checksum
      // perform XOR on 6 bytes to check for errors
      if (Rx_Buff.ESPdata[7] == (Rx_Buff.ESPdata[1] ^ Rx_Buff.ESPdata[2] ^ Rx_Buff.ESPdata[3] ^ Rx_Buff.ESPdata[4] ^ Rx_Buff.ESPdata[5] ^ Rx_Buff.ESPdata[6])) {
        // data has passed the checksum test
        RxWiFi[0] = Rx_Buff.ESPdata[1];
        RxWiFi[1] = Rx_Buff.ESPdata[2];
        RxWiFi[2] = Rx_Buff.ESPdata[3];
        RxWiFi[3] = Rx_Buff.ESPdata[4];
        RxWiFi[4] = Rx_Buff.ESPdata[5];
        RxWiFi[5] = Rx_Buff.ESPdata[6];

        // if Nunchuk data strip off the JX,JY and CZ values
        RxRec = true; // indicate a valid frame
        if (WiiType == 'N') {
          JoyX = RxWiFi[0];   // joystick X 0-255
          JoyY = RxWiFi[1];   // joystick Y 0-255
          CZ = RxWiFi[5] & 3; // CZ buttons, C = 2/0, Z = 1/0, active low
        } else {
          // not Nunchuk so set defaults
          JoyX = 128; JoyY = 128; CZ = 3;
        }
      } else {
        WiFiRxErr++;
      }
      Rx_Task = 0;  // reset the task pointer to lok at another block if needed
      Rx_Pnt += 8;  // increase the data pointer beyond this block
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0;} // reached end of buffer
      break;

    case 2: // load text data as if from a serial port
      // read characters up to the end of the buffer
      Rx_Pnt++;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0; break;} // reached end of buffer
      keyVal = Rx_Buff.ESPdata[Rx_Pnt]; PrintTgt = 1; decodeKey(keyVal);
      Serial.write(keyVal);
      break;

    case 3: // load binary data block
      // data length is specified in second byte
      Rx_Pnt++;
      // at the moment just jump over it, binary trasnfer is not needed
      Rx_Pnt += Rx_Buff.ESPdata[Rx_Pnt] + 1;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; break;} // reached end of buffer
      break;
  }
}

// ---------------------------------------------------------------------------------

void readWii() {
  // called every 20ms, but wireless transceiver only transmits every 40ms
  // considers data received from the Wii Nunchuk controller over WiFi and 
  // manages all movements demanded by the controller
  if (readWiiCall) {return;}
  readWiiCall = true; // prevent re-entry of this function
  
  if (RxRec) {
  //##############################################################################
  //
  //  WiFi Nunchuk data received
  //
  //##############################################################################
    RxRec = false; overRun = 5;
//    PrintTx += "JX" + String(JoyX) + " JY" + String(JoyY) + " CZ" + String(CZ) + "\n";
//    Serial.print("CZ "); Serial.println(CZ);
    if (!WiFiEn) {
  //##############################################################################
  //
  //  WiFi is currently disabled, listening
  //
  //##############################################################################
      // we have not received enough 'C' buttons to make the system active
      // the 'C' button alone must be held in to activate the Wii WiFi link
      if (CZ == 1) {
        // C button only is being pressed
        WiFiCntC += 4;
//        Serial.println("WiFiCnt = " + String(WiFiCntC));
        if (WiFiCntC > 60) {
          // going active, user has held in 'C' button for sufficient time
          Serial.println("WiFiEn = true");
          WiFiCntC = 64; WiFiEn = true; C_Dn = 8;
          LEDMthTask = -1;  // reset mouth LED tasks
          if (MainTask != 0) {SetMainTask(0);}
          if (!AtRest) {SetMainTask(98); LEDMthTask = -1;} // move to squat (default)
          // set defaults here
          HeadTask = 10;
          Gear = 1;
        } else {
          // still counting C buttons
          LEDMthVal = WiFiCntC/10; LEDMthTask = 10;
        }
      } else {
        // C button not pressed or released early so check for a inadequate count
        if (WiFiCntC > 0) {
          // reduce the C-button counter, which has to be sustained to enable WiFi mode
          WiFiCntC--; LEDMthVal = 1 + (WiFiCntC/10);
          if (WiFiCntC < 1) {LEDMthTask = -1;} else {LEDMthTask = 10;}
        }
      }
    } else {
      
  //##############################################################################
  //
  //  WiFi is enabled
  //
  //##############################################################################
      // WiFi enabled so check for power down
      if (CZ == 0) {
        // both C and Z buttons are being pressed so move towards powering down
        if (WiFiCntC > 0) {
          WiFiCntC -= 5;
  //        Serial.println("WiFiCnt = " + String(WiFiCntC));
          if (WiFiCntC <= 0) {
            // power-down condition satisfied, so go to rest
            Serial.println("WiFi disabled!");
            WiFiCntC = 0; disableWiFi();
            LEDTaskPnt = -1;  // eyes OFF
            if (!AtRest) {SetMainTask(98); LEDMthTask = -1;} // move to squat
            // set defaults here
          } else {
            // still counting down CZ buttons
            LEDMthVal = WiFiCntC/10; LEDMthTask = 10;
          }
        }
      } else {
        // Check CZ buttons to see if a 'Gear' change is wanted
        if ((CZ & 2) == 0) {
          // C button is pressed, assume gear change
          C_Dn++; if (C_Dn == 2) {
            C_Dn = 8; if (Gear < 5) {Gear++;}
            if (LEDTaskPnt != 99) {LEDNextTask = LEDTaskPnt;}
            LEDTaskPnt = -5; // display 'Gear' colour in the eyes
          }
//          Serial.print("Gear "); Serial.println(Gear);
        } else {C_Dn = 0;} // C button released
        // Check Z button. It has two functions. Short press change gear, but if
        // held in it changes the mode of the joysticks
        if ((CZ & 1) == 0) {
          // Z button is pressed
          Z_Dn++; if (Z_Dn > 12) {Z_Dn = 50;} // Z_Dn capped at 50, held for >= 2 seconds
//          Serial.print("Z_Dn "); Serial.println(Z_Dn);
        } else {
          if ((Z_Dn > 0) && (Z_Dn <= 10)) {
            // held down for less than 400 ms @ 40ms Rx cycle
            if (Gear > 1) {Gear--;}  // short press so change gear if possible
            if (LEDTaskPnt != 99) {LEDNextTask = LEDTaskPnt;}
            LEDTaskPnt = -5; // display 'Gear' colour in the eyes
//            Serial.print("Gear "); Serial.println(Gear);
          } Z_Dn = 0; // Z button released
        }
//        Serial.print("Z_Dn "); Serial.println(Z_Dn);

        if (WiFiCntC < 60) {
          // C-button counter is not at full strength, assume WiFi off released prematurely
          WiFiCntC++; LEDMthVal = 1 + (WiFiCntC/10);
          if (WiFiCntC == 60) {LEDMthTask = -1;} else {LEDMthTask = 10;}
        }
      }

      
  //##############################################################################
  //
  //  Wii joystick demands
  //
  //##############################################################################
      // now respond to joystick demands. We use a state machine to control how
      // the joystick demands are responded to. But first we resolve the JoyX and
      // JoyY values into a strength vector JoyV, and a direction JoyD
      // when not centred a counter is used to record how long the joystick is off centre
//      Serial.print("X"); Serial.print(JoyX); Serial.print("  Y"); Serial.println(JoyY);
      
      JoyXC = JoyX - 128; JoyYC = JoyY - 128;
      if ((abs(JoyXC) < 12) && (abs(JoyYC) < 12)) {
        // joystick is in centre deadband region
        JoyV = 0; JoyD = 0; // clear vector and direction values
      } else {
        // calculate vector outside of centre region
        JoyV = (int)sqrt((sq((float)JoyXC) + sq((float)JoyYC)));
        JoyV = min(JoyV,127); // cap max value, which can reach 181
        if (JoyXC == 0) {JoyXC = 1;}  // prevent divide by zero
        if (JoyYC == 0) {JoyYC = 1;}
        
        JoyD = 0; // reset direction pointer to default
        if (JoyYC >= 12) {
          // in upper hemisphere
          if (abs(JoyYC/JoyXC) >= 2) {JoyD = 1; JoyD1++;} // pointing N
          else if (abs(JoyXC/JoyYC) >= 2) {
            if (JoyXC > 0) {JoyD = 3;}                    // pointing E
            else {JoyD = 7;}                              // pointing W
          } else {
            if (JoyXC > 0) {JoyD = 2;}                    // pointing NE
            else {JoyD = 8;}                              // pointing NW
          }
        } else if (JoyYC > -12) {
          // in middle area
          if (JoyXC >= 12) {JoyD = 3;}                    // pointing E
          else {JoyD = 7;}                                // Pointing W
        } else {
          // in lower hemisphere
          if (abs(JoyYC/JoyXC) >= 2) {JoyD = 5;}          // pointing S
          else if (abs(JoyXC/JoyYC) >= 2) {
            if (JoyXC > 0) {JoyD = 3;}                    // pointing E
            else {JoyD = 7;}                              // pointing W
          } else {
            if (JoyXC > 0) {JoyD = 4;}                    // pointing SE
            else {JoyD = 6;}                              // pointing SW
          }
        }
      }
//      Serial.print("JoyV "); Serial.print(JoyV);
//      Serial.print("   JoyD "); Serial.println(JoyD);
//      PrintTx += String(JoyD) + "\t" + String(JoyV) + "\n"; JoyD = 0; // print joystick vectors but don't move
//      if (MoveState > 0) {PrintTx += String(Speed) + "\n";}
      
      
  //##############################################################################
  //
  //  Respond to Wii controller
  //
  //##############################################################################
//      Serial.println("JoyD " + String(JoyD) + "  MoveState " + String(JoyD));
      switch(MoveState) {
        case 0:
          // default state, droid will be in ready state or rest state
          // if in Ready state return to Stand after a delay to save power
          if (Ready) {ReadyCnt--; if (ReadyCnt < 1) {MoveToStandFromAny();}}
          
          // look for a joystick demand to initiate movement
          if (Z_Dn < 12) {
            // 'Z' button is not held down
            // if not stood up we only respond to joystick presses in key directions
            if (AtRest) {
//              Serial.println(JoyD1);
              if ((JoyD == 0) && (JoyD1 > 0)) {
                if (JoyD1 > 15) {MoveToStandFromAny();}
                else {MoveToSitLowFromRest();}
              }
            } else if (SitLow) {
              if ((JoyD == 0) && (JoyD1 > 0)) {MoveToSitHighFromLow();}
            } else if (SitHigh) {
              if ((JoyD == 0) && (JoyD1 > 0)) {MoveToStandFromAny();}
            } else {
              // stood up so respond to joystick commands
            if (JoyD == 1) {MoveFwd();}
//            else if (JoyD == 2) {MoveFwdRht();}
            else if (JoyD == 3) {MoveTurnRight();}
            else if (JoyD == 5) {MoveBkd();}
            else if (JoyD == 7) {MoveTurnLeft();}
//            else if (JoyD == 8) {MoveFwdLft();}
            }
          } else {
            // 'Z' button is held down
            if (Ready) {
                if (JoyD == 1) {MoveToSquatExercise();}
              else if (JoyD == 3) {MoveStepRight();}
              else if (JoyD == 5) {MoveToSitHighAndClap();}
              else if (JoyD == 7) {MoveStepLeft();}
            } else if (Stand) {MoveToReady(1); loopWhileWalk();}
          } break;
        case 1: // moving forward
          MoveFwd(); break;
        case 2: // moving forward bearing to the right
          MoveFwdRht(); break;
        case 3: // turning right
          MoveTurnRight(); break;
        case 5: // moving backward
          MoveBkd(); break;
        case 7: // turning left
          MoveTurnLeft(); break;
        case 8: // moving forward bearing to the left
          MoveFwdLft(); break;
        case 13: // stepping right
          MoveStepRight(); break;
        case 17: // stepping left
          MoveStepLeft(); break;
        case 99:
          // wait for joystick to centralise
          if (JoyD == 0) {
            MoveState = 0; Trans = !Trans; LEDMthTask = -1;
            Serial.println("Joystick released");
          } break;
      }
      // if joystick centred then clear the direction counter
      if (JoyD == 0) {JoyD1 = 0;}
      if (MoveState > 0) {
        // responding to joystick so reset rest counter
        ReadyCnt = ReadyCntDef;
        WiFiRestCnt = WiFiRestDef;
      }
    }
  } else {
  //##############################################################################
  //
  //  No WiFi Nunchuk received
  //
  //##############################################################################
    if (WiFiEn) {
      // droid is enabled for Wii WiFi control
      // maintain Wii enabled counter
      if ((WiFiCntC < 64) && (CZ != 0)) {
        WiFiCntC++; LEDMthVal = WiFiCntC/10;
        if (WiFiCntC == 64) {LEDMthTask = -1;} else {LEDMthTask = 10;}
      }
      
      // if in MoveState == 20 complete the move first, otherwise do these checks
      if (MoveState != 20) {
        // if moving ensure that demands are still being received, if not goto ready
        if (MoveState != 0) {
          overRun --;
//          if (overRun < 1) {MoveState = 0; MoveToReady(1); loopWhileWalk(); LEDMthTask = -1;}
        }
        // perform auto-rest function if no demands
        if (WiFiRestCnt > 0) {WiFiRestCnt--;}
        if (WiFiRestCnt == 1) {
          // if not at rest go there
          if (!AtRest) {MoveToStandFromAny();}
        }
      }
//      Serial.println(WiFiRestCnt);
    }
  }

  if (WiFiEn) {
  //##############################################################################
  //
  //  WiFi background tasks when WiFi is enabled
  //
  //##############################################################################
    // only do these functions if WiFi is enabled
    if (MoveState == 20) {
      // long un-demanded moves like bowing are watched until they complete
      if (Walk < 1) {MoveState = 0; HeadTask = 10; LEDMthTask = -1;}
    }
    
    if (MoveState == -1) {
      // at the end of a joystick move we want to set things back to normal
      MoveToReady(1); loopWhileWalk(); LEDMthTask = -1; SetSpeedDefaults();
      MoveState = 0;
    }
  }
  
  readWiiCall = false; // re-allow calls to this function
//  Serial.print("CntZ "); Serial.println(WiFiCntZ);
//   Serial.println(MoveState);
}

// --------------------------------------------------------------------------------

void RxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Rx buffer of length zlen
  Rx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    Rx_Chk = Rx_Chk ^ Rx_Buff.ESPdata[zP];
  }
}

// ---------------------------------------------------------------------------------

void SetMainTask(int zM) {
  // called to change current task and set initial conditions
  Detect = false;   // target detection flag; true detected
  DetectCnt = 0;    // target detection counter
  ESC0 = false;     // ESCape flag, normally if button SW0 is pressed
  ESC1 = false;     // ESCape flag, normally if button SW1 is pressed
  Head = 0;         // terminate any head movement
  HeadTask = 0;     // head task pointer
  LEDMthTask = -1;  // reset mouth tasks and clear LEDs
  LEDMthVal = 0;    // clear variable
  LEDSubTask = 0;   // start subtasks from 0
  if (LTOF_On > 0) {VL53L0X_OFF();}
  mainEn = true;    // enable main tasks
  mainSubCnt = 0;   // zero the counter
  mainSubTask = 0;  // start subtasks from 0
  MoveState = 0;    // reset the move state
  Pause = false;    // pause flag causes movement to be put on hold
  setBalanceOFF();  // turn OFF self-balancing in case it was ON
  Walk = 0;         // terminate any leg movement
  WalkMem = 0;      // pointer to next move row in Mem[] to be read, 0 == do nothing
  WalkMode = 0;     // set the mode when moving using Mem[] targets, 0 == once, 1 = continuous
  
  LEDTaskPnt = 2;
  switch(zM) {
//    case 1: LEDTaskPnt = 2; break;
  }
  MainTask = zM;
  if (GR == 0) {PrintTx += "\nMainTask = " + String(MainTask) + "\n";}
  synchLoopTimers(); // ensure sync has not been lost
}

// ---------------------------------------------------------------------------------

void SetPCAop() {
  // sets all PCA channels to there default values
  for (int zC=0; zC < 16; zC++) {
    I2Cpwm.setPWM(zC, 0, SRV[zC]);  // if OFFtime == ONtime the op never goes HIGH
  }
}

// ---------------------------------------------------------------------------------

void SetServoOE(int zVal) {
  // manually set the servo board OE to HIGH or LOW
  // 0 - servos ON with outpus set
  // 1 - servoos OFF with all PWM set to zero
  if (zVal != 1) {
    // enabling PCA9685 outputs
    servoOE = 0;
    digitalWrite(Pin_OE, LOW); // enable all PWM outputs
//    Serial.println("OE=LO [ON]");
  } else {
    // disabling PCA9685 outputs. As this is asynchronous there is a risk that this
    // action will shorten one or more PWM pulses, so we remove them first
    servoOE = 1; ClearPCAop();
//    delay(10 + (1000/int(pwmFreq)));  // wait longer than one PWM cycle
    digitalWrite(Pin_OE, HIGH); // disable all PWM outputs
//    Serial.println("OE=HI [OFF]");
    synchLoopTimers();
  }
}

// ---------------------------------------------------------------------------------

void SetServoPW(int zCh, int zPWM) {
  // set the pulse width of a specified channel on the PCA9685 board
  // format: setPWM(channel,on,off)
  // setting 'on' = 0 means that the pulse 'off' can be 0 - 4095 (12-bits)
  I2Cpwm.setPWM(zCh, 0, zPWM);
  if (servoOE ==1) {SetServoOE(0);}  // if board is not enabled, enable it
}

// ---------------------------------------------------------------------------------

void SetSpeedDefaults() {
  // called to set default speed range values
  SpdMax = 100;           // maximum speed when moving
  SpdMid = 50;            // middle speed when moving
  SpdMin = 20;            // minimum speed when moving
  SpdTop = SpdMid;        // current top speed when moving
  Speed = SpdMin;         // default speed value
}

// ---------------------------------------------------------------------------------

void shiftBatData() {
  // shift battery data so that it is a rolling average
  for (int zB = 0; zB < 59; zB++) {BatData[zB] = BatData[zB + 1];}
}

// ---------------------------------------------------------------------------------

void VL53L0X_OFF () {
  // disable the VL53L0X for reading ranges
  digitalWrite(XSHUT,LOW);  // VL53L0X reset is active LOW
  LTOF_On = 0;              // > 0 when laser ranging is enabled, default is 0 == OFF
}

// ---------------------------------------------------------------------------------

void VL53L0X_ON () {
  // enable the VL53L0X for reading ranges
  digitalWrite(XSHUT,HIGH); // enable the VL53L0X
  VL53L0X_init(true); delay(10);
  VL53L0X_setTimeout(500);

  // Start continuous back-to-back mode (take readings as
  // fast as possible).  To use continuous timed mode
  // instead, provide a desired inter-measurement period in
  // ms (e.g. sensor.startContinuous(100)).
  VL53L0X_startContinuous(); delay(10);
  LTOF_On = 1; // restart the measurement statemachine
}

// --------------------------------------------------------------------------------




  //###############################################################################
  //
  //  WiFi Code
  //
  //###############################################################################

void WiFiClearRxBuff(int zNum) {
  // clears the contents of the receive buffer
  for (int zP = 0; zP < zNum;zP++) {Rx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void WiFiClearTxBuff(int zNum) {
  // clears the contents of the transmit buffer
  for (int zP = 0; zP < zNum;zP++) {Tx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void WiFiDisconnect() {
  // enter the WiFi disconnected state
  WiFiConnected = false;
  WiFiClearRxBuff(WiFiTx_len);
  WiFiClearTxBuff(WiFiTx_len);
  WiFiTryCnt = 0; // reset the connection retry counter
  WiFiTx_CB = 1;  // set ready to send flag
  Serial.println("WiFiDisconnected!");
}

// --------------------------------------------------------------------------------

void WiFiTryToConnect() {
  // try to connect with the Wii Transceiver by sending devices name
  Tx_Buff.ESPdata[0] = 'R';
  Tx_Buff.ESPdata[1] = 'o';
  Tx_Buff.ESPdata[2] = 'b';
  Tx_Buff.ESPdata[3] = 'o';
  Tx_Buff.ESPdata[4] = 'D';
  Tx_Buff.ESPdata[5] = 'o';
  Tx_Buff.ESPdata[6] = 'g';
  WiFiTx_len = 7;
//  t0 = micros();
  esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
  WiFiTryNum++; // count the total number of attempts
//  Serial.println("WiFiTryToConnect...");
}

// --------------------------------------------------------------------------------

void WiFiTxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Tx buffer of length zlen
  WiFiTx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    WiFiTx_Chk = WiFiTx_Chk ^ Tx_Buff.ESPdata[zP];
  }
}

// ---------------------------------------------------------------------------------
